""" Sub-package containing policy routines
"""

from aizynthfinder.context.policy.expansion_strategies import (
    ExpansionStrategy,
    MultiExpansionStrategy,
    TemplateBasedDirectExpansionStrategy,
    TemplateBasedExpansionStrategy,
)
from aizynthfinder.context.policy.filter_strategies import (
    BondFilter,
    FilterStrategy,
    QuickKerasFilter,
    ReactantsCountFilter,
)
from aizynthfinder.context.policy.policies import ExpansionPolicy, FilterPolicy
from aizynthfinder.utils.exceptions import PolicyException



""" Module containing a class that is the main interface the retrosynthesis tool.
"""
from __future__ import annotations

import time
from collections import defaultdict
from typing import TYPE_CHECKING

from tqdm import tqdm

from aizynthfinder.analysis import (
    RouteCollection,
    RouteSelectionArguments,
    TreeAnalysis,
)
from aizynthfinder.chem import FixedRetroReaction, Molecule, TreeMolecule
from aizynthfinder.context.config import Configuration
from aizynthfinder.context.policy import BondFilter
from aizynthfinder.context.scoring import BrokenBondsScorer, CombinedScorer
from aizynthfinder.reactiontree import ReactionTreeFromExpansion
from aizynthfinder.search.andor_trees import AndOrSearchTreeBase
from aizynthfinder.search.mcts import MctsSearchTree
from aizynthfinder.utils.exceptions import MoleculeException
from aizynthfinder.utils.loading import load_dynamic_class

# This must be imported first to setup logging for rdkit, tensorflow etc
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.utils.type_utils import (
        Callable,
        Dict,
        List,
        Optional,
        StrDict,
        Tuple,
        Union,
    )


class AiZynthFinder:
    """
    Public API to the aizynthfinder tool

    If instantiated with the path to a yaml file or dictionary of settings
    the stocks and policy networks are loaded directly.
    Otherwise, the user is responsible for loading them prior to
    executing the tree search.

    :ivar config: the configuration of the search
    :ivar expansion_policy: the expansion policy model
    :ivar filter_policy: the filter policy model
    :ivar stock: the stock
    :ivar scorers: the loaded scores
    :ivar tree: the search tree
    :ivar analysis: the tree analysis
    :ivar routes: the top-ranked routes
    :ivar search_stats: statistics of the latest search

    :param configfile: the path to yaml file with configuration (has priority over configdict), defaults to None
    :param configdict: the config as a dictionary source, defaults to None
    """

    def __init__(
        self, configfile: Optional[str] = None, configdict: Optional[StrDict] = None
    ) -> None:
        self._logger = logger()

        if configfile:
            self.config = Configuration.from_file(configfile)
        elif configdict:
            self.config = Configuration.from_dict(configdict)
        else:
            self.config = Configuration()

        self.expansion_policy = self.config.expansion_policy
        self.filter_policy = self.config.filter_policy
        self.stock = self.config.stock
        self.scorers = self.config.scorers
        self.tree: Optional[Union[MctsSearchTree, AndOrSearchTreeBase]] = None
        self._target_mol: Optional[Molecule] = None
        self.search_stats: StrDict = dict()
        self.routes = RouteCollection([])
        self.analysis: Optional[TreeAnalysis] = None
        self._num_objectives = len(
            self.config.search.algorithm_config.get("search_rewards", [])
        )

    @property
    def target_smiles(self) -> str:
        """The SMILES representation of the molecule to predict routes on."""
        if not self._target_mol:
            return ""
        return self._target_mol.smiles

    @target_smiles.setter
    def target_smiles(self, smiles: str) -> None:
        self.target_mol = Molecule(smiles=smiles)

    @property
    def target_mol(self) -> Optional[Molecule]:
        """The molecule to predict routes on"""
        return self._target_mol

    @target_mol.setter
    def target_mol(self, mol: Molecule) -> None:
        self.tree = None
        self._target_mol = mol

    def build_routes(
        self,
        selection: Optional[RouteSelectionArguments] = None,
        scorer: Optional[Union[str, List[str]]] = None,
    ) -> None:
        """
        Build reaction routes

        This is necessary to call after the tree search has completed in order
        to extract results from the tree search.

        :param selection: the selection criteria for the routes
        :param scorer: a reference to the object used to score the nodes, can be a list
        :raises ValueError: if the search tree not initialized
        """
        self.analysis = self._setup_analysis(scorer=scorer)
        config_selection = RouteSelectionArguments(
            nmin=self.config.post_processing.min_routes,
            nmax=self.config.post_processing.max_routes,
            return_all=self.config.post_processing.all_routes,
        )
        self.routes = RouteCollection.from_analysis(
            self.analysis, selection or config_selection
        )

    def extract_statistics(self) -> StrDict:
        """Extracts tree statistics as a dictionary"""
        if not self.analysis:
            return {}
        stats = {
            "target": self.target_smiles,
            "search_time": self.search_stats["time"],
            "first_solution_time": self.search_stats.get("first_solution_time", 0),
            "first_solution_iteration": self.search_stats.get(
                "first_solution_iteration", 0
            ),
        }
        stats.update(self.analysis.tree_statistics())
        return stats

    def prepare_tree(self) -> None:
        """
        Setup the tree for searching

        :raises ValueError: if the target molecule was not set
        """
        if not self.target_mol:
            raise ValueError("No target molecule set")

        try:
            self.target_mol.sanitize()
        except MoleculeException:
            raise ValueError("Target molecule unsanitizable")

        self.stock.reset_exclusion_list()
        if (
            self.config.search.exclude_target_from_stock
            and self.target_mol in self.stock
        ):
            self.stock.exclude(self.target_mol)
            self._logger.debug("Excluding the target compound from the stock")

        if self.config.search.break_bonds or self.config.search.freeze_bonds:
            self._setup_focussed_bonds(self.target_mol)

        self._setup_search_tree()
        self.analysis = None
        self.routes = RouteCollection([])
        self.filter_policy.reset_cache()
        self.expansion_policy.reset_cache()

    def stock_info(self) -> StrDict:
        """
        Return the stock availability for all leaf nodes in all collected reaction trees

        The key of the return dictionary will be the SMILES string of the leaves,
        and the value will be the stock availability

        :return: the collected stock information.
        """
        if not self.analysis:
            return {}
        _stock_info = {}
        for tree in self.routes.reaction_trees:
            for leaf in tree.leafs():
                if leaf.smiles not in _stock_info:
                    _stock_info[leaf.smiles] = self.stock.availability_list(leaf)
        return _stock_info

    def tree_search(self, show_progress: bool = False) -> float:
        """
        Perform the actual tree search

        :param show_progress: if True, shows a progress bar
        :return: the time past in seconds
        """
        if not self.tree:
            self.prepare_tree()
        # This is for type checking, prepare_tree is creating it.
        assert self.tree is not None
        self.search_stats = {"returned_first": False, "iterations": 0}

        time0 = time.time()
        i = 1
        self._logger.debug("Starting search")
        time_past = time.time() - time0

        if show_progress:
            pbar = tqdm(total=self.config.search.iteration_limit, leave=False)

        while (
            time_past < self.config.search.time_limit
            and i <= self.config.search.iteration_limit
        ):
            if show_progress:
                pbar.update(1)
            self.search_stats["iterations"] += 1

            try:
                is_solved = self.tree.one_iteration()
            except StopIteration:
                break

            if is_solved and "first_solution_time" not in self.search_stats:
                self.search_stats["first_solution_time"] = time.time() - time0
                self.search_stats["first_solution_iteration"] = i

            if self.config.search.return_first and is_solved:
                self._logger.debug("Found first solved route")
                self.search_stats["returned_first"] = True
                break
            i = i + 1
            time_past = time.time() - time0

        if show_progress:
            pbar.close()
        time_past = time.time() - time0
        self._logger.debug("Search completed")
        self.search_stats["time"] = time_past
        return time_past

    def _setup_focussed_bonds(self, target_mol: Molecule) -> None:
        """
        Setup multi-objective scoring function with 'broken bonds'-scorer and
        add 'frozen bonds'-filter to filter policy.

        :param target_mol: the target molecule.
        """
        target_mol = TreeMolecule(smiles=target_mol.smiles, parent=None)

        bond_filter_key = "__finder_bond_filter"
        if self.config.search.freeze_bonds:
            if not target_mol.has_all_focussed_bonds(self.config.search.freeze_bonds):
                raise ValueError("Bonds in 'freeze_bond' must exist in target molecule")
            bond_filter = BondFilter(bond_filter_key, self.config)
            self.filter_policy.load(bond_filter)
            self.filter_policy.select(bond_filter_key, append=True)
        elif (
            self.filter_policy.selection
            and bond_filter_key in self.filter_policy.selection
        ):
            self.filter_policy.deselect(bond_filter_key)

        search_rewards = self.config.search.algorithm_config.get("search_rewards")
        if not search_rewards:
            return

        if self.config.search.break_bonds and "broken bonds" in search_rewards:
            if not target_mol.has_all_focussed_bonds(self.config.search.break_bonds):
                raise ValueError("Bonds in 'break_bonds' must exist in target molecule")
            self.scorers.load(BrokenBondsScorer(self.config))
            self._num_objectives = len(search_rewards)

    def _setup_search_tree(self) -> None:
        self._logger.debug(f"Defining tree root:  {self.target_smiles}")
        if self.config.search.algorithm.lower() == "mcts":
            self.tree = MctsSearchTree(
                root_smiles=self.target_smiles, config=self.config
            )
        else:
            cls = load_dynamic_class(self.config.search.algorithm)
            self.tree = cls(root_smiles=self.target_smiles, config=self.config)

    def _setup_analysis(
        self,
        scorer: Optional[Union[str, List[str]]],
    ) -> TreeAnalysis:
        """Configure TreeAnalysis

        :param scorer: a reference to the object used to score the nodes, can be a list
        :returns: the configured TreeAnalysis
        :raises ValueError: if the search tree not initialized
        """
        if not self.tree:
            raise ValueError("Search tree not initialized")

        if scorer is None:
            scorer_names = self.config.post_processing.route_scorers
            # If not defined, use the same scorer as the search rewards
            if not scorer_names:
                search_rewards = self.config.search.algorithm_config.get(
                    "search_rewards"
                )
                scorer_names = search_rewards if search_rewards else ["state score"]

        elif isinstance(scorer, str):
            scorer_names = [scorer]
        else:
            scorer_names = list(scorer)

        if "broken bonds" in scorer_names:
            # Add broken bonds scorer if required
            self.scorers.load(BrokenBondsScorer(self.config))

        scorers = [self.scorers[name] for name in scorer_names]

        if self.config.post_processing.scorer_weights:
            scorers = [
                CombinedScorer(
                    self.config,
                    scorer_names,
                    self.config.post_processing.scorer_weights,
                )
            ]

        return TreeAnalysis(self.tree, scorers)


class AiZynthExpander:
    """
    Public API to the AiZynthFinder expansion and filter policies

    If instantiated with the path to a yaml file or dictionary of settings
    the stocks and policy networks are loaded directly.
    Otherwise, the user is responsible for loading them prior to
    executing the tree search.

    :ivar config: the configuration of the search
    :ivar expansion_policy: the expansion policy model
    :ivar filter_policy: the filter policy model

    :param configfile: the path to yaml file with configuration (has priority over configdict), defaults to None
    :param configdict: the config as a dictionary source, defaults to None
    """

    def __init__(
        self, configfile: Optional[str] = None, configdict: Optional[StrDict] = None
    ) -> None:
        self._logger = logger()

        if configfile:
            self.config = Configuration.from_file(configfile)
        elif configdict:
            self.config = Configuration.from_dict(configdict)
        else:
            self.config = Configuration()

        self.expansion_policy = self.config.expansion_policy
        self.filter_policy = self.config.filter_policy
        self.stats: StrDict = {}

    def do_expansion(
        self,
        smiles: str,
        return_n: int = 5,
        filter_func: Optional[Callable[[RetroReaction], bool]] = None,
    ) -> List[Tuple[FixedRetroReaction, ...]]:
        """
        Do the expansion of the given molecule returning a list of
        reaction tuples. Each tuple in the list contains reactions
        producing the same reactants. Hence, nested structure of the
        return value is way to group reactions.

        If filter policy is setup, the probability of the reactions are
        added as metadata to the reaction.

        The additional filter functions makes it possible to do customized
        filtering. The callable should take as only argument a `RetroReaction`
        object and return True if the reaction can be kept or False if it should
        be removed.

        :param smiles: the SMILES string of the target molecule
        :param return_n: the length of the return list
        :param filter_func: an additional filter function
        :return: the grouped reactions
        """
        self.stats = {"non-applicable": 0}

        mol = TreeMolecule(parent=None, smiles=smiles)
        actions, _ = self.expansion_policy.get_actions([mol])
        results: Dict[Tuple[str, ...], List[FixedRetroReaction]] = defaultdict(list)
        for action in actions:
            reactants = action.reactants
            if not reactants:
                self.stats["non-applicable"] += 1
                continue
            if filter_func and not filter_func(action):
                continue
            for name in self.filter_policy.selection or []:
                if hasattr(self.filter_policy[name], "feasibility"):
                    _, feasibility_prob = self.filter_policy[name].feasibility(action)
                    action.metadata["feasibility"] = float(feasibility_prob)
                    break
            action.metadata["expansion_rank"] = len(results) + 1
            unique_key = tuple(sorted(mol.inchi_key for mol in reactants[0]))
            if unique_key not in results and len(results) >= return_n:
                continue
            rxn = next(ReactionTreeFromExpansion(action).tree.reactions())  # type: ignore
            results[unique_key].append(rxn)
        return [tuple(reactions) for reactions in results.values()]



""" Module containing routines for returning package paths
"""
import os


def package_path() -> str:
    """Return the path to the package"""
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def data_path() -> str:
    """Return the path to the ``data`` directory of the package"""
    return os.path.join(package_path(), "data")



""" Module containing a classes representation various tree nodes
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from aizynthfinder.chem import TreeMolecule
from aizynthfinder.chem.serialization import deserialize_action, serialize_action
from aizynthfinder.search.andor_trees import TreeNodeMixin

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.chem.serialization import (
        MoleculeDeserializer,
        MoleculeSerializer,
    )
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import List, Optional, Sequence, Set, StrDict


class MoleculeNode(TreeNodeMixin):
    """
    An OR node representing a molecule

    :ivar expandable: if True, this node is part of the frontier
    :ivar mol: the molecule represented by the node
    :ivar in_stock: if True the molecule is in stock and hence should not be expanded
    :ivar parent: the parent of the node

    :param mol: the molecule to be represented by the node
    :param config: the configuration of the search
    :param parent: the parent of the node, optional
    """

    def __init__(
        self,
        mol: TreeMolecule,
        config: Configuration,
        parent: Optional[ReactionNode] = None,
    ) -> None:
        self.mol = mol
        self._config = config
        self.in_stock = mol in config.stock
        self.parent = parent

        self._children: List[ReactionNode] = []
        # Makes it unexpandable if we have reached maximum depth
        self.expandable = self.mol.transform < self._config.search.max_transforms

        if self.in_stock:
            self.expandable = False

    @classmethod
    def create_root(cls, smiles: str, config: Configuration) -> "MoleculeNode":
        """
        Create a root node for a tree using a SMILES.

        :param smiles: the SMILES representation of the root state
        :param config: settings of the tree search algorithm
        :return: the created node
        """
        mol = TreeMolecule(parent=None, transform=0, smiles=smiles)
        return MoleculeNode(mol=mol, config=config)

    @classmethod
    def from_dict(
        cls,
        dict_: StrDict,
        config: Configuration,
        molecules: MoleculeDeserializer,
        parent: Optional[ReactionNode] = None,
    ) -> "MoleculeNode":
        """
        Create a new node from a dictionary, i.e. deserialization

        :param dict_: the serialized node
        :param config: settings of the tree search algorithm
        :param molecules: the deserialized molecules
        :param parent: the parent node
        :return: a deserialized node
        """
        mol = molecules.get_tree_molecules([dict_["mol"]])[0]
        node = MoleculeNode(mol, config, parent)
        node.expandable = dict_["expandable"]
        node.children = [
            ReactionNode.from_dict(child, config, molecules, parent=node)
            for child in dict_["children"]
        ]
        return node

    @property  # type: ignore
    def children(self) -> List[ReactionNode]:  # type: ignore
        """Gives the reaction children nodes"""
        return self._children

    @children.setter
    def children(self, value: List[ReactionNode]) -> None:
        self._children = value

    @property
    def prop(self) -> StrDict:
        return {"solved": self.in_stock, "mol": self.mol}

    def add_stub(self, reaction: RetroReaction) -> Sequence[MoleculeNode]:
        """
        Add a stub / sub-tree to this node

        :param reaction: the reaction creating the stub
        :return: list of all newly added molecular nodes
        """
        reactants = reaction.reactants[reaction.index]
        if not reactants:
            return []

        ancestors = self.ancestors()
        for mol in reactants:
            if mol in ancestors:
                return []

        rxn_node = ReactionNode.create_stub(
            reaction=reaction, parent=self, config=self._config
        )
        self._children.append(rxn_node)

        return rxn_node.children

    def ancestors(self) -> Set[TreeMolecule]:
        """
        Return the ancestors of this node

        :return: the ancestors
        :rtype: set
        """
        if not self.parent:
            return {self.mol}

        ancestors = self.parent.parent.ancestors()
        ancestors.add(self.mol)
        return ancestors

    def serialize(self, molecule_store: MoleculeSerializer) -> StrDict:
        """
        Serialize the node object to a dictionary

        :param molecule_store: the serialized molecules
        :return: the serialized node
        """
        dict_: StrDict = {"expandable": self.expandable}
        dict_["mol"] = molecule_store[self.mol]
        dict_["children"] = [child.serialize(molecule_store) for child in self.children]
        return dict_


class ReactionNode(TreeNodeMixin):
    """
    An AND node representing a reaction

    :ivar parent: the parent of the node
    :ivar reaction: the reaction represented by the node

    :param cost: the cost of the reaction
    :param reaction: the reaction to be represented by the node
    :param parent: the parent of the node
    """

    def __init__(self, reaction: RetroReaction, parent: MoleculeNode) -> None:
        self.parent = parent
        self.reaction = reaction

        self._children: List[MoleculeNode] = []

    @classmethod
    def create_stub(
        cls,
        reaction: RetroReaction,
        parent: MoleculeNode,
        config: Configuration,
    ) -> ReactionNode:
        """
        Create a ReactionNode and creates all the MoleculeNode objects
        that are the children of the node.

        :param reaction: the reaction to be represented by the node
        :param parent: the parent of the node
        :param config: the configuration of the search tree
        """
        node = cls(reaction, parent)
        reactants = reaction.reactants[reaction.index]
        node.children = [
            MoleculeNode(mol=mol, config=config, parent=node) for mol in reactants
        ]
        return node

    @classmethod
    def from_dict(
        cls,
        dict_: StrDict,
        config: Configuration,
        molecules: MoleculeDeserializer,
        parent: MoleculeNode,
    ) -> ReactionNode:
        """
        Create a new node from a dictionary, i.e. deserialization

        :param dict_: the serialized node
        :param config: the configuration of the tree search
        :param molecules: the deserialized molecules
        :param parent: the parent node
        :return: a deserialized node
        """
        reaction = deserialize_action(dict_["reaction"], molecules)
        node = cls(reaction, parent)

        node.children = [
            MoleculeNode.from_dict(child, config, molecules, parent=node)
            for child in dict_["children"]
        ]
        return node

    @property  # type: ignore
    def children(self) -> List[MoleculeNode]:  # type: ignore
        """Gives the molecule children nodes"""
        return self._children

    @children.setter
    def children(self, value: List[MoleculeNode]) -> None:
        self._children = value

    @property
    def prop(self) -> StrDict:
        return {"solved": False, "reaction": self.reaction}

    def serialize(self, molecule_store: MoleculeSerializer) -> StrDict:
        """
        Serialize the node object to a dictionary

        :param molecule_store: the serialized molecules
        :return: the serialized node
        """
        dict_ = {
            "reaction": serialize_action(self.reaction, molecule_store),
            "children": [child.serialize(molecule_store) for child in self.children],
        }
        return dict_



""" Module containing classes to deal with Reactions.
"""
from __future__ import annotations

import abc
import hashlib
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
from rdchiral import main as rdc

try:
    from rdchiral.bonds import get_atoms_across_double_bonds
    from rdchiral.initialization import BondDirOpposite
except ImportError:
    RDCHIRAL_CPP = True
else:
    RDCHIRAL_CPP = False
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.rdchem import BondDir, BondStereo, ChiralType

from aizynthfinder.chem.mol import (
    Molecule,
    MoleculeException,
    TreeMolecule,
)
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.chem.mol import UniqueMolecule
    from aizynthfinder.utils.type_utils import (
        Any,
        Iterable,
        List,
        Optional,
        RdReaction,
        Set,
        StrDict,
        Tuple,
        Union,
    )

if RDCHIRAL_CPP:
    logger().warning(
        "WARNING: C++ version of RDChiral is supported, but with limited functionality"
    )


class _ReactionInterfaceMixin:
    """
    Mixin class to define a common interface for all reaction class

    The methods `_products_getter` and `_reactants_getter` needs to be implemented by subclasses
    """

    def fingerprint(
        self, radius: int, nbits: Optional[int] = None, chiral: bool = False
    ) -> np.ndarray:
        """
        Returns a difference fingerprint

        :param radius: the radius of the fingerprint
        :param nbits: the length of the fingerprint. If not given it will use RDKit default, defaults to None
        :param chiral: if True, include chirality information
        :return: the fingerprint
        """
        product_fp = sum(
            mol.fingerprint(radius, nbits, chiral) for mol in self._products_getter()  # type: ignore
        )
        reactants_fp = sum(
            mol.fingerprint(radius, nbits, chiral) for mol in self._reactants_getter()  # type: ignore
        )
        return reactants_fp - product_fp  # type: ignore

    def hash_list(self) -> List[str]:
        """
        Return all the products and reactants as hashed SMILES

        :return: the hashes of the SMILES string
        """
        mols = self.reaction_smiles().replace(".", ">>").split(">>")
        return [hashlib.sha224(mol.encode("utf8")).hexdigest() for mol in mols]

    def hash_key(self) -> str:
        """
        Return a code that can be use to identify the reaction

        :return: the hash code
        """
        reactants = sorted([mol.inchi_key for mol in self._reactants_getter()])  # type: ignore
        products = sorted([mol.inchi_key for mol in self._products_getter()])  # type: ignore
        hash_ = hashlib.sha224()
        for item in reactants + [">>"] + products:
            hash_.update(item.encode())
        return hash_.hexdigest()

    def rd_reaction_from_smiles(self) -> RdReaction:
        """
        The reaction as a RDkit reaction object but created from the reaction smiles
        instead of the SMARTS of the template.

        :return: the reaction object
        """
        return AllChem.ReactionFromSmarts(self.reaction_smiles(), useSmiles=True)

    def reaction_smiles(self) -> str:
        """
        Get the reaction SMILES, i.e. the SMILES of the reactants and products joined together

        :return: the SMILES
        """
        reactants = ".".join(mol.smiles for mol in self._reactants_getter())  # type: ignore
        products = ".".join(mol.smiles for mol in self._products_getter())  # type: ignore
        return f"{reactants}>>{products}"


class RetroReaction(abc.ABC, _ReactionInterfaceMixin):
    """
    A retrosynthesis reaction. Only a single molecule is the reactant.

    This is an abstract class and child classes needs to implement the `_apply` and `_make_smiles` functions
    that should create the reactants molecule objects and the reaction SMILES representation, respectively.

    :ivar mol: the TreeMolecule object that this reaction is applied to
    :ivar index: a unique index of this reaction,
                 to count for the fact that a reaction can produce more than one outcome
    :ivar metadata: meta data associated with the reaction

    :param mol: the molecule
    :param index: the index, defaults to 0
    :param metadata: some meta data
    :params kwargs: any extra parameters for child classes
    """

    _required_kwargs: List[str] = []

    def __init__(
        self,
        mol: TreeMolecule,
        index: int = 0,
        metadata: Optional[StrDict] = None,
        **kwargs: Any,
    ) -> None:
        if any(name not in kwargs for name in self._required_kwargs):
            raise KeyError(
                f"A {self.__class__.__name__} class needs to be initiated "
                f"with keyword arguments: {', '.join(self._required_kwargs)}"
            )
        self.mol = mol
        self.index = index
        self.metadata: StrDict = metadata or {}
        self._reactants: Optional[Tuple[Tuple[TreeMolecule, ...], ...]] = None
        self._smiles: Optional[str] = None
        self._kwargs: StrDict = kwargs

    @classmethod
    def from_serialization(
        cls, init_args: StrDict, reactants: List[List[TreeMolecule]]
    ) -> "RetroReaction":
        """
        Create an object from a serialization. It does
        1) instantiate an object using the `init_args` and
        2) set the reactants to a tuple-form of `reactants

        :param init_args: the arguments passed to the `__init__` method
        :param reactants: the reactants
        :return: the deserialized object
        """
        obj = cls(**init_args)
        obj._reactants = tuple(tuple(mol for mol in lst_) for lst_ in reactants)
        return obj

    def __str__(self) -> str:
        return f"reaction on molecule {self.mol.smiles}"

    @property
    def reactants(self) -> Tuple[Tuple[TreeMolecule, ...], ...]:
        """
        Returns the reactant molecules.
        Apply the reaction if necessary.

        :return: the products of the reaction
        """
        if not self._reactants:
            self._reactants = self._apply()
        return self._reactants

    @property
    def smiles(self) -> str:
        """
        The reaction as a SMILES

        :return: the SMILES
        """
        if self._smiles is None:
            try:
                self._smiles = self._make_smiles()
            except ValueError:
                self._smiles = ""  # noqa
        return self._smiles

    @property
    def unqueried(self) -> bool:
        """
        Return True if the reactants has never been retrieved
        """
        return self._reactants is None

    def copy(self, index: Optional[int] = None) -> "RetroReaction":
        """
        Shallow copy of this instance.

        :param index: new index, defaults to None
        :return: the copy
        """
        # pylint: disable=protected-access
        index = index if index is not None else self.index
        new_reaction = self.__class__(
            self.mol, index, dict(self.metadata), **self._kwargs
        )
        new_reaction._reactants = tuple(mol_list for mol_list in self._reactants or [])
        new_reaction._smiles = self._smiles
        return new_reaction

    def mapped_reaction_smiles(self) -> str:
        """
        Get the mapped reaction SMILES if it exists
        :return: the SMILES
        """
        reactants = self.mol.mapped_smiles
        products = ".".join(mol.mapped_smiles for mol in self._products_getter())
        return reactants + ">>" + products

    def to_dict(self) -> StrDict:
        """
        Return the retro reaction as dictionary
        This dictionary is not suitable for serialization, but is used by other serialization routines
        The elements of the dictionary can be used to instantiate a new reaction object
        """
        return {
            "mol": self.mol,
            "index": self.index,
            "metadata": dict(self.metadata),
        }

    @abc.abstractmethod
    def _apply(self) -> Tuple[Tuple[TreeMolecule, ...], ...]:
        pass

    @abc.abstractmethod
    def _make_smiles(self) -> str:
        pass

    def _products_getter(self) -> Tuple[TreeMolecule, ...]:
        return self.reactants[self.index]

    def _reactants_getter(self) -> List[TreeMolecule]:
        return [self.mol]

    @staticmethod
    def _update_unmapped_atom_num(mol: TreeMolecule, exclude_nums: Set[int]) -> None:
        mapped_nums = {num for num in mol.mapping_to_index.keys() if 0 < num < 900}
        offset = max(mapped_nums) + 1 if mapped_nums else 1
        for atom in mol.mapped_mol.GetAtoms():
            if 0 < atom.GetAtomMapNum() < 900:
                continue
            while offset in exclude_nums:
                offset += 1
            atom.SetAtomMapNum(offset)
            exclude_nums.add(offset)


class TemplatedRetroReaction(RetroReaction):
    """
    A retrosynthesis reaction that uses a reaction SMARTS and RDChiral to produce reactant molecules.
    The SMILES representation of the reaction is the SMARTS (modified by RDKit)

    :param mol: the molecule
    :param index: the index, defaults to 0
    :param metadata: some meta data
    :param smarts: a string representing the template
    """

    _required_kwargs = ["smarts"]

    def __init__(
        self,
        mol: TreeMolecule,
        index: int = 0,
        metadata: Optional[StrDict] = None,
        **kwargs: Any,
    ):
        super().__init__(mol, index, metadata, **kwargs)
        self.smarts: str = kwargs["smarts"]
        self._use_rdchiral: bool = kwargs.get("use_rdchiral", True)
        self._rd_reaction: Optional[RdReaction] = None

    def __str__(self) -> str:
        return (
            f"retro reaction from template {self.smarts} on molecule {self.mol.smiles}"
        )

    @property
    def rd_reaction(self) -> RdReaction:
        """Return the RDKit reaction created from the SMART"""
        if self._rd_reaction is None:
            self._rd_reaction = AllChem.ReactionFromSmarts(self.smarts)
        return self._rd_reaction

    def to_dict(self) -> StrDict:
        dict_ = super().to_dict()
        dict_["smarts"] = self.smarts
        return dict_

    def _apply(self) -> Tuple[Tuple[TreeMolecule, ...], ...]:
        if self._use_rdchiral:
            return self._apply_with_rdchiral()
        return self._apply_with_rdkit()

    def _apply_with_rdchiral(self) -> Tuple[Tuple[TreeMolecule, ...], ...]:
        """
        Apply a reactions smarts to a molecule and return the products (reactants for retro templates)
        Will try to sanitize the reactants, and if that fails it will not return that molecule
        """
        reaction = rdc.rdchiralReaction(self.smarts)
        rct = _RdChiralProductWrapper(self.mol)
        try:
            reactants = rdc.rdchiralRun(reaction, rct, keep_mapnums=True)
        except RuntimeError as err:
            logger().debug(
                f"Runtime error in RDChiral with template {self.smarts} on {self.mol.smiles}\n{err}"
            )
            reactants = []
        except KeyError as err:
            logger().debug(
                f"Index error in RDChiral with template {self.smarts} on {self.mol.mapped_smiles}\n{err}"
            )
            reactants = []

        # Turning rdchiral outcome into rdkit tuple of tuples to maintain compatibility
        outcomes = []
        for reactant_str in reactants:
            smiles_list = reactant_str.split(".")
            exclude_nums = set(self.mol.mapping_to_index.keys())
            update_func = partial(
                self._update_unmapped_atom_num, exclude_nums=exclude_nums
            )
            try:
                rct_objs = tuple(
                    TreeMolecule(
                        parent=self.mol,
                        smiles=smi,
                        sanitize=True,
                        mapping_update_callback=update_func,
                    )
                    for smi in smiles_list
                )
            except MoleculeException:
                pass
            else:
                outcomes.append(rct_objs)
        self._reactants = tuple(outcomes)

        return self._reactants

    def _apply_with_rdkit(self) -> Tuple[Tuple[TreeMolecule, ...], ...]:
        rxn = AllChem.ReactionFromSmarts(self.smarts)
        try:
            reactants_list = rxn.RunReactants([self.mol.mapped_mol])
        except:  # pylint: disable=bare-except
            reactants_list = []

        outcomes = []
        for reactants in reactants_list:
            exclude_nums = set(self.mol.mapping_to_index.keys())
            update_func = partial(self._inherit_atom_mapping, exclude_nums=exclude_nums)
            try:
                mols = tuple(
                    TreeMolecule(
                        parent=self.mol,
                        rd_mol=mol,
                        sanitize=True,
                        mapping_update_callback=update_func,
                    )
                    for mol in reactants
                )
            except MoleculeException:
                pass
            else:
                outcomes.append(mols)
        self._reactants = tuple(outcomes)

        return self._reactants

    def _make_smiles(self):
        return AllChem.ReactionToSmiles(self.rd_reaction)

    def _inherit_atom_mapping(self, mol: TreeMolecule, exclude_nums: Set[int]) -> None:
        """
        Update the internal atom mapping dictionary by inspecting the `reaction_atom_idx`
        property of the atoms and comparing it with the parent-molecule.

        This is used for child molecules created by RDKit reaction application.
        RDChiral takes care of this automatically.
        """
        if mol.parent is None:
            return

        for atom in mol.mapped_mol.GetAtoms():
            if not atom.HasProp("react_atom_idx"):
                continue
            index = atom.GetProp("react_atom_idx")
            mapping = mol.parent.index_to_mapping.get(int(index))
            if mapping:
                atom.SetAtomMapNum(mapping)

        self._update_unmapped_atom_num(mol, exclude_nums)


class SmilesBasedRetroReaction(RetroReaction):
    """
    A retrosynthesis reaction where the SMILES of the reactants are given on initiation

    The SMILES representation of the reaction is the reaction SMILES

    :param mol: the molecule
    :param index: the index, defaults to 0
    :param metadata: some meta data
    :param reactants_str: a dot-separated string of reactant SMILES strings
    """

    _required_kwargs = ["reactants_str"]

    def __init__(
        self,
        mol: TreeMolecule,
        index: int = 0,
        metadata: Optional[StrDict] = None,
        **kwargs: Any,
    ):
        super().__init__(mol, index, metadata, **kwargs)
        self.reactants_str: str = kwargs["reactants_str"]
        self._mapped_prod_smiles = kwargs.get("mapped_prod_smiles")

    def __str__(self) -> str:
        return (
            f"retro reaction on molecule {self.mol.smiles} giving {self.reactants_str}"
        )

    def to_dict(self) -> StrDict:
        dict_ = super().to_dict()
        dict_["reactants_str"] = self.reactants_str
        dict_["mapped_prod_smiles"] = self._mapped_prod_smiles
        return dict_

    def _apply(self) -> Tuple[Tuple[TreeMolecule, ...], ...]:
        outcomes = []
        smiles_list = self.reactants_str.split(".")

        exclude_nums = set(self.mol.mapping_to_index.keys())
        update_func = partial(self._remap, exclude_nums=exclude_nums)
        try:
            rct = tuple(
                TreeMolecule(
                    parent=self.mol,
                    smiles=smi,
                    sanitize=True,
                    mapping_update_callback=update_func,
                )
                for smi in smiles_list
            )
        except MoleculeException:
            pass
        else:
            outcomes.append(rct)
        self._reactants = tuple(outcomes)

        return self._reactants

    def _remap(self, mol: TreeMolecule, exclude_nums: Set[int]) -> None:
        """Find the mapping between parent and child and then re-map the child molecule"""
        if not self._mapped_prod_smiles:
            self._update_unmapped_atom_num(mol, exclude_nums)
            return

        parent_remapping = {}
        pmol = Molecule(smiles=self._mapped_prod_smiles, sanitize=True)
        for atom_idx1, atom_idx2 in enumerate(
            pmol.rd_mol.GetSubstructMatch(self.mol.mapped_mol)
        ):
            atom1 = self.mol.mapped_mol.GetAtomWithIdx(atom_idx1)
            atom2 = pmol.rd_mol.GetAtomWithIdx(atom_idx2)
            if atom1.GetAtomMapNum() > 0 and atom2.GetAtomMapNum() > 0:
                parent_remapping[atom2.GetAtomMapNum()] = atom1.GetAtomMapNum()

        for atom in mol.mapped_mol.GetAtoms():
            if atom.GetAtomMapNum() and atom.GetAtomMapNum() in parent_remapping:
                atom.SetAtomMapNum(parent_remapping[atom.GetAtomMapNum()])
            else:
                atom.SetAtomMapNum(0)

        self._update_unmapped_atom_num(mol, exclude_nums)

    def _make_smiles(self):
        rstr = ".".join(reactant.smiles for reactant in self.reactants[0])
        return f"{self.mol.smiles}>>{rstr}"


class FixedRetroReaction(_ReactionInterfaceMixin):
    """
    A retrosynthesis reaction that has the same interface as `RetroReaction`
    but it is fixed so it does not support SMARTS application or any creation of reactants.

    The reactants are set by using the `reactants` property.

    :ivar mol: the UniqueMolecule object that this reaction is applied to
    :ivar smiles: the SMILES representation of the RDKit reaction
    :ivar metadata: meta data associated with the reaction
    :ivar reactants: the reactants of this reaction

    :param mol: the molecule
    :param smiles: the SMILES of the reaction
    :param metadata: some meta data
    """

    def __init__(
        self,
        mol: UniqueMolecule,
        smiles: str = "",
        metadata: Optional[StrDict] = None,
    ) -> None:
        self.mol = mol
        self.smiles = smiles
        self.metadata = metadata or {}
        self.reactants: Tuple[Tuple[UniqueMolecule, ...], ...] = ()

    def copy(self) -> "FixedRetroReaction":
        """
        Shallow copy of this instance.

        :return: the copy
        """
        new_reaction = FixedRetroReaction(self.mol, self.smiles, self.metadata)
        new_reaction.reactants = tuple(mol_list for mol_list in self.reactants)
        return new_reaction

    def to_smiles_based_retroreaction(self) -> SmilesBasedRetroReaction:
        """
        Convert a FixedRetroReaction to a SmilesBasedRetroReaction.

        :return: the SmilesBasedRetroReaction.
        """
        if self.metadata and "mapped_reaction_smiles" in self.metadata.keys():
            mapped_reaction_smiles = self.metadata["mapped_reaction_smiles"]
        else:
            mapped_reaction_smiles = self.reaction_smiles()

        mapped_reaction_smiles = mapped_reaction_smiles.split(">>")
        product = mapped_reaction_smiles[0]
        reactants = mapped_reaction_smiles[1]

        return SmilesBasedRetroReaction(
            mol=TreeMolecule(smiles=product, parent=None),
            mapped_prod_smiles=product,
            reactants_str=reactants,
        )

    def _products_getter(self) -> Tuple[UniqueMolecule, ...]:
        return self.reactants[0]

    def _reactants_getter(self) -> List[UniqueMolecule]:
        return [self.mol]


def hash_reactions(
    reactions: Union[Iterable[RetroReaction], Iterable[FixedRetroReaction]],
    sort: bool = True,
) -> str:
    """
    Creates a hash for a list of reactions

    :param reactions: the reactions to hash
    :param sort: if True will sort all molecules, defaults to True
    :return: the hash string
    """
    hash_list = []
    for reaction in reactions:
        hash_list.extend(reaction.hash_list())
    if sort:
        hash_list.sort()
    hash_list_str = ".".join(hash_list)
    return hashlib.sha224(hash_list_str.encode("utf8")).hexdigest()


class _RdChiralProductWrapper:
    """
    Reimplementation of `rdchiralReaction`
    to preserve product molecule already created
    """

    # pylint: disable=W0106,C0103
    def __init__(self, product: TreeMolecule) -> None:
        product.sanitize()
        self.reactant_smiles = product.smiles

        # Initialize into RDKit mol
        self.reactants = Chem.Mol(product.mapped_mol.ToBinary())
        Chem.AssignStereochemistry(self.reactants, flagPossibleStereoCenters=True)
        self.reactants.UpdatePropertyCache(strict=False)

        self.atoms_r = {a.GetAtomMapNum(): a for a in self.reactants.GetAtoms()}
        self.idx_to_mapnum = lambda idx: self.reactants.GetAtomWithIdx(
            idx
        ).GetAtomMapNum()

        # Create copy of molecule without chiral information, used with
        # RDKit's naive runReactants
        self.reactants_achiral = Chem.Mol(product.rd_mol.ToBinary())
        [
            a.SetChiralTag(ChiralType.CHI_UNSPECIFIED)
            for a in self.reactants_achiral.GetAtoms()
        ]
        [
            (b.SetStereo(BondStereo.STEREONONE), b.SetBondDir(BondDir.NONE))
            for b in self.reactants_achiral.GetBonds()
        ]

        # Pre-list reactant bonds (for stitching broken products)
        self.bonds_by_mapnum = [
            (b.GetBeginAtom().GetAtomMapNum(), b.GetEndAtom().GetAtomMapNum(), b)
            for b in self.reactants.GetBonds()
        ]

        # Pre-list chiral double bonds (for copying back into outcomes/matching)
        self.bond_dirs_by_mapnum = {}
        for i, j, b in self.bonds_by_mapnum:
            if b.GetBondDir() != BondDir.NONE:
                self.bond_dirs_by_mapnum[(i, j)] = b.GetBondDir()
                self.bond_dirs_by_mapnum[(j, i)] = BondDirOpposite[b.GetBondDir()]

        # Get atoms across double bonds defined by mapnum
        self.atoms_across_double_bonds = get_atoms_across_double_bonds(self.reactants)


if RDCHIRAL_CPP:

    def _wrapper(mol):
        return rdc.rdchiralReactants(mol.mapped_smiles)

    _RdChiralProductWrapper = _wrapper  # type: ignore



"""Module containing utility functions for GUI.
"""
from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from IPython.display import HTML, display
from paretoset import paretorank

if TYPE_CHECKING:
    from aizynthfinder.analysis.routes import RouteCollection
    from aizynthfinder.utils.type_utils import List, Optional


def pareto_fronts_plot(
    routes: RouteCollection,
) -> None:
    """Plot the pareto front(s).

    :param routes: the route collection to plot as Pareto fronts
    """
    scorer_names = list(routes.scores[0].keys())
    scores = np.array(
        [[score_dict[name] for name in scorer_names] for score_dict in routes.scores]
    )
    direction_arr = np.repeat("max", len(scorer_names))
    pareto_ranks = paretorank(scores, sense=direction_arr, distinct=False)

    pareto_fronts = pd.DataFrame(scores, columns=scorer_names)
    pareto_fronts.loc[:, "pareto_rank"] = pareto_ranks
    pareto_fronts.loc[:, "route"] = np.arange(1, scores.shape[0] + 1)
    pareto_fronts_unique = pareto_fronts.drop_duplicates(scorer_names)
    # Apply the default theme
    sns.set_theme()
    fig = sns.relplot(
        data=pareto_fronts_unique,
        x=scorer_names[0],
        y=scorer_names[1],
        hue="pareto_rank",
        kind="line",
        markers=True,
    )
    fig.set(
        xlabel=f"{scorer_names[0]}",
        ylabel=f"{scorer_names[1]}",
        title="Pareto Fronts",
    )

    objectives2solutions = defaultdict(list)
    for _, row in pareto_fronts.iterrows():
        x_val = row[scorer_names[0]]
        y_val = row[scorer_names[1]]
        objectives2solutions[(x_val, y_val)].append(int(row["route"]))

    # Add route label on pareto line(s)
    for _, row in pareto_fronts_unique.iterrows():
        x_val = row[scorer_names[0]]
        y_val = row[scorer_names[1]]
        point_val = f"Option {_values_to_string(objectives2solutions[(x_val, y_val)])}"
        plt.text(x=x_val, y=y_val, s=point_val, size=10)
    plt.show()


def route_display(
    index: Optional[int], routes: RouteCollection, output_widget: widgets.Output
) -> None:
    """
    Display a route with auxillary information in a widget

    :param index: the index of the route to display
    :param routes: the route collection
    :param output_widget: the widget to display the route on
    """
    if index is None or routes is None or index >= len(routes):
        return

    route = routes[index]
    state = route["node"].state
    status = "Solved" if state.is_solved else "Not Solved"

    output_widget.clear_output()
    with output_widget:
        display(HTML(f"<H2>{status}"))
        table_content = "".join(
            f"<tr><td>{name}</td><td>{score:.4f}</td></tr>"
            if isinstance(score, (float, int))
            else f"<tr><td>{name}</td><td>({', '.join(f'{val:.4f}' for val in score)})</td></tr>"
            for name, score in route["all_score"].items()
        )
        display(HTML(f"<table>{table_content}</table>"))
        display(HTML("<H2>Compounds to Procure"))
        display(state.to_image())
        display(HTML("<H2>Steps"))
        display(routes[index]["image"])


def _values_to_string(vals: List[int]) -> str:
    """
    Given a list of integers, produce a nice-looking string

    Example:
         >> _values_to_string([1,2,3,10,11,19])
         "1-3, 10-11, 19"
    """
    groups = []
    start = end = vals[0]
    for prev_val, val in zip(vals[:-1], vals[1:]):
        if prev_val == val - 1:
            end += 1
        else:
            groups.append((start, end))
            start = end = val
    groups.append((start, end))

    group_strs = []
    for start, end in groups:
        if start == end:
            group_strs.append(str(start))
        else:
            group_strs.append(f"{start}-{end}")
        if len(group_strs) % 5 == 0:
            group_strs[-1] = "\n" + group_strs[-1]
    return ", ".join(group_strs)



""" Module containing diverse math functions, including neural network-related functions. """

import numpy as np
from aizynthfinder.utils.type_utils import Callable


# pylint: disable=invalid-name
def dense_layer_forward_pass(
    x: np.ndarray, weights: np.ndarray, bias: np.ndarray, activation: Callable
) -> np.ndarray:
    """
    Forward pass through a dense neural network layer.
    :param x: layer input
    :param weights: layer weights
    :param bias: layer bias
    :param activation: layer activation function
    :return: the layer output
    """
    x = np.matmul(x, weights) + bias
    return activation(x)


# pylint: disable=invalid-name
def rectified_linear_unit(x: np.ndarray) -> np.ndarray:
    """ReLU activation function"""
    return x * (x > 0)


# pylint: disable=invalid-name
def sigmoid(x: np.ndarray) -> np.ndarray:
    """Sigmoid activation function"""
    return 1 / (1 + np.exp(-x))


# pylint: disable=invalid-name
def softmax(x: np.ndarray) -> np.ndarray:
    """Compute softmax values for each sets of scores in x."""
    return np.exp(x) / np.sum(np.exp(x), axis=0)



""" Module containing a GUI extension for clustering
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from IPython.display import display
from ipywidgets import BoundedIntText, Button, HBox, Label, Output, Tab

try:
    import matplotlib.pylab as plt
    from route_distances.clustering import ClusteringHelper
    from scipy.cluster.hierarchy import dendrogram
except ImportError:
    raise ImportError(
        "Clustering is not supported by this installation."
        " Please install aizynthfinder with extras dependencies."
    )


if TYPE_CHECKING:
    from aizynthfinder.analysis import RouteCollection
    from aizynthfinder.interfaces.aizynthapp import AiZynthApp
    from aizynthfinder.utils.type_utils import StrDict


class ClusteringGui:
    """
    GUI extension to cluster routes

    :param routes: the routes to cluster
    :param content: what to cluster on
    """

    def __init__(self, routes: RouteCollection, content: str = "both"):
        self._routes = routes
        self._routes.distance_matrix(content=content, recreate=True)
        self._input: StrDict = dict()
        self._output: StrDict = dict()
        self._buttons: StrDict = dict()
        self._create_dendrogram()
        self._create_input()
        self._create_output()

    @classmethod
    def from_app(cls, app: AiZynthApp, content: str = "both"):
        """
        Helper function to create a GUI from a GUI app interface

        :param app: the app to extract the routes from
        :param content: what to cluster on
        :return: the GUI object
        """
        if app.finder.routes is None:
            raise ValueError("Cannot initialize GUI from no routes")
        return ClusteringGui(app.finder.routes, content)

    def _create_dendrogram(self) -> None:
        dend_out = Output(
            layout={"width": "99%", "height": "310px", "overflow_y": "auto"}
        )
        with dend_out:
            print("This is the hierarchy of the routes")
            fig = plt.Figure()
            dendrogram(
                ClusteringHelper(self._routes.distance_matrix()).linkage_matrix(),
                color_threshold=0.0,
                labels=np.arange(1, len(self._routes) + 1),
                ax=fig.gca(),
            )
            fig.gca().set_xlabel("Route")
            fig.gca().set_ylabel("Distance")
            display(fig)
        display(dend_out)

    def _create_input(self) -> None:
        self._input["number_clusters"] = BoundedIntText(
            continuous_update=True,
            min=1,
            max=len(self._routes) - 1,
            layout={"width": "80px"},
        )
        self._buttons["cluster"] = Button(description="Cluster")
        self._buttons["cluster"].on_click(self._on_cluster_button_clicked)
        box = HBox(
            [
                Label("Number of clusters to make"),
                self._input["number_clusters"],
                self._buttons["cluster"],
            ]
        )
        display(box)
        help_out = Output()
        with help_out:
            print(
                "Optimization is carried out if the number of given clusters are less than 2"
            )
        display(help_out)

    def _create_output(self) -> None:
        self._output["clusters"] = Tab()
        display(self._output["clusters"])

    def _on_cluster_button_clicked(self, _) -> None:
        self._buttons["cluster"].enabled = False
        self._routes.cluster(self._input["number_clusters"].value)
        self._buttons["cluster"].enabled = True

        outputs = []
        for i, cluster in enumerate(self._routes.clusters or []):
            output = Output(
                layout={
                    "border": "1px solid silver",
                    "width": "99%",
                    "height": "500px",
                    "overflow_y": "auto",
                }
            )
            with output:
                for image in cluster.images:
                    print(f"Route {self._routes.images.index(image)+1}")
                    display(image)
            outputs.append(output)
            self._output["clusters"].set_title(i, f"Cluster {i+1}")
        self._output["clusters"].children = outputs



""" Module containing classes to store and manipulate collections of synthetic routes.
"""
from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import numpy as np

try:
    from route_distances.clustering import ClusteringHelper
    from route_distances.route_distances import route_distances_calculator
except ImportError:
    pass

from aizynthfinder.analysis.utils import CombinedReactionTrees, RouteSelectionArguments
from aizynthfinder.reactiontree import SUPPORT_DISTANCES, ReactionTree
from aizynthfinder.search.mcts import MctsNode, MctsSearchTree

if TYPE_CHECKING:
    from aizynthfinder.analysis import TreeAnalysis
    from aizynthfinder.context.scoring import Scorer
    from aizynthfinder.utils.type_utils import (
        Any,
        Dict,
        Optional,
        PilImage,
        Sequence,
        StrDict,
        Union,
    )


class RouteCollection:
    """
    Holds a collections of reaction routes.

    If can be the top scored nodes, their scores and
    the reaction trees created from them.
    It can also be a cluster of such routes.

    The class has the functionality to compute collective results
    for the different routes such as images.

    Properties of individual route can be obtained with simple indexing.

    .. code-block::

        route0 = collection[0]

    :ivar all_scores: all the computed scores for the routes
    :ivar nodes: the top-ranked MCTS-like nodes
    :ivar scores: initial scores of top-ranked nodes or routes
    :ivar reaction_trees: the reaction trees created from the top-ranked nodes
    :ivar clusters: the created clusters from the collection
    :ivar route_metadata: the metadata of the reaction trees

    :param reaction_trees: the trees to base the collection on
    """

    def __init__(self, reaction_trees: Sequence[ReactionTree], **kwargs) -> None:
        self._routes: Sequence[StrDict] = [{} for _ in range(len(reaction_trees))]
        self.reaction_trees = reaction_trees
        self._update_route_dict(reaction_trees, "reaction_tree")
        self.route_metadata = [rt.metadata for rt in reaction_trees]
        self._update_route_dict(self.route_metadata, "route_metadata")

        self.nodes = self._unpack_kwarg_with_default("nodes", None, **kwargs)
        self.scores = self._unpack_kwarg_with_default("scores", dict, **kwargs)
        self.all_scores = self._unpack_kwarg_with_default("all_scores", dict, **kwargs)

        self._dicts: Optional[Sequence[StrDict]] = self._unpack_kwarg("dicts", **kwargs)
        self._images: Optional[Sequence[Optional[PilImage]]] = self._unpack_kwarg(
            "images", **kwargs
        )
        self._jsons: Optional[Sequence[str]] = self._unpack_kwarg("jsons", **kwargs)
        self.clusters: Optional[Sequence[RouteCollection]] = self._unpack_kwarg(
            "clusters", **kwargs
        )
        self._distance_matrix: Dict[str, np.ndarray] = {}
        self._combined_reaction_trees: Optional[CombinedReactionTrees] = None

    @classmethod
    def from_analysis(
        cls,
        analysis: TreeAnalysis,
        selection: Optional[RouteSelectionArguments] = None,
    ) -> "RouteCollection":
        """
        Create a collection from a tree analysis.

        :param analysis: the tree analysis to use
        :param selection: selection criteria for the routes
        :return: the created collection
        """
        items, scores = analysis.sort(selection)
        all_scores = copy.deepcopy(scores)
        kwargs: Dict[str, Any] = {"scores": scores, "all_scores": all_scores}
        if isinstance(analysis.search_tree, MctsSearchTree):
            kwargs["nodes"] = items
            reaction_trees = [
                from_node.to_reaction_tree()
                for from_node in items
                if isinstance(from_node, MctsNode)
            ]
        else:
            reaction_trees = items  # type: ignore

        return cls(reaction_trees, **kwargs)

    def __getitem__(self, index: int) -> StrDict:
        if index < 0 or index >= len(self):
            raise IndexError("Index out of range")
        return self._routes[index]

    def __len__(self) -> int:
        return len(self.reaction_trees)

    @property
    def dicts(self) -> Sequence[StrDict]:
        """Returns a list of dictionary representation of the routes"""
        if self._dicts is None:
            self._dicts = self.make_dicts()
        return self._dicts

    @property
    def images(self) -> Sequence[Optional[PilImage]]:
        """Returns a list of pictoral representation of the routes"""
        if self._images is None:
            self._images = self.make_images()
        assert self._images is not None
        return self._images

    @property
    def jsons(self) -> Sequence[str]:
        """Returns a list of JSON string representation of the routes"""
        if self._jsons is None:
            self._jsons = self.make_jsons()
        return self._jsons

    def cluster(
        self,
        n_clusters: int,
        max_clusters: int = 5,
        distances_model: str = "ted",
        **kwargs: Any,
    ) -> np.ndarray:
        """
        Cluster the route collection into a number of clusters.

        Additional arguments to the distance or clustering algorithm
        can be passed in as key-word arguments.

        When `distances_model` is "lstm", a key-word argument `model_path` needs to be given
        when `distances_model` is "ted", two optional key-word arguments `timeout` and `content`
        can be given.

        If the number of reaction trees are less than 3, no clustering will be performed

        :param n_clusters: the desired number of clusters, if less than 2 triggers optimization
        :param max_clusters: the maximum number of clusters to consider
        :param distances_model: can be ted or lstm and determines how the route distances are computed
        :return: the cluster labels
        """
        if not SUPPORT_DISTANCES:
            raise ValueError(
                "Clustering is not supported by this installation."
                " Please install aizynthfinder with extras dependencies."
            )

        if len(self.reaction_trees) < 3:
            return np.asarray([])
        dist_kwargs = {
            "content": kwargs.pop("content", "both"),
            "timeout": kwargs.pop("timeout", None),
            "model_path": kwargs.pop("model_path", None),
        }
        try:
            distances = self.distance_matrix(model=distances_model, **dist_kwargs)
        except ValueError:
            return np.asarray([])

        labels = ClusteringHelper.cluster(
            distances,
            n_clusters,
            max_clusters=max_clusters,
            **kwargs,
        )
        self._make_clusters(labels)
        return labels

    def combined_reaction_trees(self, recreate: bool = False) -> CombinedReactionTrees:
        """
        Return an object that combines all the reaction tree into a single reaction tree graph

        :param recreate: if False will return a cached object if available, defaults to False
        :return: the combined trees
        """
        if not self._combined_reaction_trees or recreate:
            self._combined_reaction_trees = CombinedReactionTrees(self.reaction_trees)
        return self._combined_reaction_trees

    def compute_scores(self, *scorers: Scorer) -> None:
        """
        Compute new scores for all routes in this collection.
        They can then be accessed with the ``all_scores`` attribute.
        """
        if self.nodes[0]:
            list_ = self.nodes
        else:
            list_ = self.reaction_trees

        for scorer in scorers:
            for idx, score in enumerate(scorer(list_)):  # type: ignore
                self.all_scores[idx][repr(scorer)] = score
        self._update_route_dict(self.all_scores, "all_score")

    def dict_with_extra(
        self, include_scores=False, include_metadata=False
    ) -> Sequence[StrDict]:
        """
        Return the routes as dictionaries with optionally all scores and
        all metadata added to the root (target) node.

        :return: the routes as dictionaries
        """
        dicts = []
        for dict_, scores, metadata in zip(
            self.dicts, self.all_scores, self.route_metadata
        ):
            dicts.append(dict(dict_))
            if include_scores:
                dicts[-1]["scores"] = dict(scores)
            if include_metadata:
                dicts[-1]["metadata"] = dict(metadata)
        return dicts

    def dict_with_scores(self) -> Sequence[StrDict]:
        """
        Return the routes as dictionaries with all scores added
        to the root (target) node.

        :return: the routes as dictionaries
        """
        return self.dict_with_extra(include_scores=True)

    def distance_matrix(
        self, recreate: bool = False, model: str = "ted", **kwargs: Any
    ) -> np.ndarray:
        """
        Compute the distance matrix between each pair of reaction trees

        All key-word arguments are passed along to the `route_distance_calculator`
        function from the `route_distances` package.

        When `model` is "lstm", a key-word argument `model_path` needs to be given
        when `model` is "ted", two optional key-word arguments `timeout` and `content`
        can be given.

        :param recreate: if False, use a cached one if available
        :param model: the type of model to use "ted" or "lstm"
        :return: the square distance matrix
        """
        if not SUPPORT_DISTANCES:
            raise ValueError(
                "Distance calculations are not supported by this installation."
                " Please install aizynthfinder with extras dependencies."
            )

        if model == "lstm" and not kwargs.get("model_path"):
            raise KeyError(
                "Need to provide 'model_path' argument when using LSTM model for computing distances"
            )
        content = kwargs.get("content", "both")
        cache_key = kwargs.get("model_path", "") if model == "lstm" else content
        if self._distance_matrix.get(cache_key) is not None and not recreate:
            return self._distance_matrix[cache_key]
        calculator = route_distances_calculator(model, **kwargs)
        distances = calculator(self.dicts)
        self._distance_matrix[cache_key] = distances
        return distances

    def make_dicts(self) -> Sequence[StrDict]:
        """Convert all reaction trees to dictionaries"""
        self._dicts = [tree.to_dict() for tree in self.reaction_trees]
        self._update_route_dict(self._dicts, "dict")
        return self._dicts

    def make_images(self) -> Sequence[Optional[PilImage]]:
        """Convert all reaction trees to images"""

        self._images = []
        for tree in self.reaction_trees:
            try:
                img = tree.to_image()
            except ValueError:
                self._images.append(None)
            else:
                self._images.append(img)
        self._update_route_dict(self._images, "image")
        return self._images

    def make_jsons(self) -> Sequence[str]:
        """Convert all reaction trees to JSON strings"""
        self._jsons = [tree.to_json() for tree in self.reaction_trees]
        self._update_route_dict(self._jsons, "json")
        return self._jsons

    def rescore(self, scorer: Scorer) -> None:
        """
        Rescore the routes in the collection, and thereby re-order them.

        This will replace the ``scores`` attribute, and update the ``all_scores``
        attribute with another entry.

        :param scorer: the scorer to use
        """
        if self.nodes[0]:
            self.nodes, self.scores, sortidx = scorer.sort(self.nodes)
            self.reaction_trees = [self.reaction_trees[idx] for idx in sortidx]
        else:
            self.reaction_trees, self.scores, sortidx = scorer.sort(self.reaction_trees)
        self._routes = [self._routes[idx] for idx in sortidx]
        self.all_scores = [self.all_scores[idx] for idx in sortidx]
        if self._dicts:
            self._dicts = [self._dicts[idx] for idx in sortidx]
        if self._images:
            self._images = [self._images[idx] for idx in sortidx]
        if self._jsons:
            self._jsons = [self._jsons[idx] for idx in sortidx]

        for idx, score in enumerate(self.scores):
            self.all_scores[idx][repr(scorer)] = score
        self._update_route_dict(self.all_scores, "all_score")

    def _make_clusters(self, clusters: np.ndarray) -> None:
        n_clusters = max(clusters) + 1
        self.clusters = []
        for cluster in range(n_clusters):
            selection = clusters == cluster
            kwargs = {
                "reaction_trees": self._select_subset(self.reaction_trees, selection),
                "nodes": self._select_subset(self.nodes, selection),
                "scores": self._select_subset(self.scores, selection),
            }
            if self._images:
                kwargs["images"] = self._select_subset(self.images, selection)
            if self._dicts:
                kwargs["dicts"] = self._select_subset(self.dicts, selection)
            if self._jsons:
                kwargs["jsons"] = self._select_subset(self.jsons, selection)

            self.clusters.append(RouteCollection(**kwargs))

    def _unpack_kwarg(self, key: str, **kwargs: Any) -> Optional[Sequence[Any]]:
        if key not in kwargs:
            return None
        arr = kwargs[key]
        self._update_route_dict(arr, key[:-1])
        return arr

    def _unpack_kwarg_with_default(
        self, key: str, default: Any, **kwargs: Any
    ) -> Sequence[Any]:
        arr = self._unpack_kwarg(key, **kwargs)
        if arr is not None:
            return arr
        return [
            default() if callable(default) else default
            for _ in range(len(self.reaction_trees))
        ]

    def _update_route_dict(self, arr: Sequence[Any], key: str) -> None:
        for i, value in enumerate(arr):
            self._routes[i][key] = value

    @staticmethod
    def _select_subset(arr: Sequence[Any], selection: Sequence[bool]) -> Sequence[Any]:
        return [item for sel, item in zip(selection, arr) if sel]



""" Module containing a class that holds the tree search
"""
from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

import networkx as nx

from aizynthfinder.chem import MoleculeDeserializer, MoleculeSerializer
from aizynthfinder.search.mcts.node import MctsNode, ParetoMctsNode
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import List, Optional, Sequence, Union


_MODE2NODECLASS = {
    "single-objective": MctsNode,
    "weighted-sum": MctsNode,
    "multi-objective": ParetoMctsNode,
}


class MctsSearchTree:
    """
    Encapsulation of the search tree.

    :ivar root: the root node
    :ivar config: the configuration of the search tree

    :param config: settings of the tree search algorithm
    :param root_smiles: the root will be set to a node representing this molecule, defaults to None
    """

    def __init__(
        self, config: Configuration, root_smiles: Optional[str] = None
    ) -> None:
        self._logger = logger()
        self.profiling = {
            "expansion_calls": 0,
            "reactants_generations": 0,
            "iterations": 0,
        }
        self.config = config
        self.mode = self._check_mode()
        self._logger.debug(f"MCTS mode: {self.mode}")

        if root_smiles:
            self.root: Optional[MctsNode] = _MODE2NODECLASS[self.mode].create_root(
                smiles=root_smiles, tree=self, config=config
            )
        else:
            self.root = None

        self._graph: Optional[nx.DiGraph] = None

        # For backward compatibility
        if "search_reward" in config.search.algorithm_config:
            config.search.algorithm_config["search_rewards"] = [
                config.search.algorithm_config["search_reward"]
            ]
        config_rewards = config.search.algorithm_config["search_rewards"]
        self._logger.setLevel(logging.WARNING)  # Supress logging from `make_subset`
        self._logger.debug(f"Selecting reward scorers: {config_rewards}")
        self.reward_scorer = self.config.scorers.make_subset(config_rewards)
        self._logger.setLevel(logging.DEBUG)
        if self.mode == "single-objective":
            self.reward_scorer_name = config_rewards[0]

    @classmethod
    def from_json(cls, filename: str, config: Configuration) -> "MctsSearchTree":
        """
        Create a new search tree by deserialization from a JSON file

        :param filename: the path to the JSON node
        :param config: the configuration of the search
        :return: a deserialized tree
        """
        tree = MctsSearchTree(config)
        with open(filename, "r") as fileobj:
            dict_ = json.load(fileobj)
        mol_deser = MoleculeDeserializer(dict_["molecules"])
        tree.root = _MODE2NODECLASS[tree.mode].from_dict(
            dict_["tree"], tree, config, mol_deser
        )
        return tree

    def backpropagate(self, from_node: MctsNode) -> None:
        """
        Backpropagate the value estimate and update all nodes from a
        given node all the way to the root.

        :param from_node: the end node of the route to update
        """
        value_estimate = self.compute_reward(from_node)

        current = from_node
        while current is not self.root:
            parent = current.parent
            # For mypy, parent should never by None unless current is the root
            assert parent is not None
            parent.backpropagate(current, value_estimate)  # type: ignore
            current = parent

    def compute_reward(self, node: MctsNode) -> Union[float, Sequence[float]]:
        """
        Compute the reward of a node in the search tree, using one of

            1. Single-objective
            2. Multi-objective
            3. Weighted-sum of multiple specified rewards

        :param node: the node to compute the reward for
        :returns: the value from the scorer(s)
        """
        if self.mode == "single-objective":
            return self.reward_scorer[self.reward_scorer_name](node)

        if self.mode == "multi-objective":
            return self.reward_scorer.score_vector(node)

        return self.reward_scorer.weighted_score(
            node, self.config.search.algorithm_config["search_rewards_weights"]
        )

    def graph(self, recreate: bool = False) -> nx.DiGraph:
        """
        Construct a directed graph object with the nodes as
        vertices and the actions as edges attribute "action".

        :param recreate: if True will construct the graph even though it is cached, defaults to False
        :return: the graph object
        :raises ValueError: if the tree is not defined
        """
        if not self.root:
            raise ValueError("Root of search tree is not defined ")

        if not recreate and self._graph:
            return self._graph

        def add_node(node):
            self._graph.add_edge(node.parent, node, action=node.parent[node]["action"])
            for grandchild in node.children:
                add_node(grandchild)

        self._graph = nx.DiGraph()
        # Always add the root
        self._graph.add_node(self.root)
        for child in self.root.children:
            add_node(child)
        return self._graph

    def nodes(self) -> List[MctsNode]:
        """Return all the nodes in the search tree"""
        return list(self.graph())

    def one_iteration(self) -> bool:
        """
        Perform one iteration of
            1. Selection
            2. Expansion
            3. Rollout
            4. Backpropagation

        :return: if a solution was found
        """
        self.profiling["iterations"] += 1
        leaf = self.select_leaf()
        leaf.expand()
        while not leaf.is_terminal():
            child = leaf.promising_child()
            if child:
                child.expand()
                leaf = child

        self.backpropagate(leaf)
        return leaf.state.is_solved

    def select_leaf(self) -> MctsNode:
        """
        Traverse the tree selecting the most promising child at
        each step until leaf node returned.

        :return: the leaf node
        :raises ValueError: if the tree is not defined
        """
        if not self.root:
            raise ValueError("Root of search tree is not defined ")

        current = self.root
        while current.is_expanded and not current.state.is_solved:
            promising_child = current.promising_child()
            # If promising_child returns None it means that the node
            # is unexpandable, and hence we should break the loop
            if promising_child:
                current = promising_child
        return current

    def serialize(self, filename: str) -> None:
        """
        Serialize the search tree to a JSON file

        :param filename: the path to the JSON file
        :raises ValueError: if the tree is not defined
        """
        if not self.root:
            raise ValueError("Root of search tree is not defined ")

        mol_ser = MoleculeSerializer()
        dict_ = {"tree": self.root.serialize(mol_ser), "molecules": mol_ser.store}
        with open(filename, "w") as fileobj:
            json.dump(dict_, fileobj, indent=2)

    def _check_mode(self) -> str:
        # if no objective weights are supplied, use multi-objective search
        # if only one objective is specified, search will in be in multi objective mode,
        #  but it is simply a vectoried version of single objective mode
        nrewards = len(self.config.search.algorithm_config["search_rewards"])
        nweights = len(self.config.search.algorithm_config["search_rewards_weights"])

        if nrewards == 1:
            mode = "single-objective"
        elif nweights == 0:
            mode = "multi-objective"
        else:
            mode = "weighted-sum"

        if mode == "weighted-sum" and nrewards != nweights:
            raise ValueError(
                "number of reward weights are expected to be the same as number of objectives,"
                f"currently have {nweights} weights and {nrewards} objectives)"
            )
        return mode



""" Module containing routine to dynamically load a class from a specification """
from __future__ import annotations

import importlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Optional


def load_dynamic_class(
    name_spec: str,
    default_module: Optional[str] = None,
    exception_cls: Any = ValueError,
) -> Any:
    """
    Load an object from a dynamic specification.

    The specification can be either:
        ClassName, in-case the module name is taken from the `default_module` argument
    or
        package_name.module_name.ClassName, in-case the module is taken as `package_name.module_name`

    :param name_spec: the class specification
    :param default_module: the default module
    :param exception_cls: the exception class to raise on exception
    :return: the loaded class
    """
    if "." not in name_spec:
        name = name_spec
        if not default_module:
            raise exception_cls(
                "Must provide default_module argument if not given in name_spec"
            )
        module_name = default_module
    else:
        module_name, name = name_spec.rsplit(".", maxsplit=1)

    try:
        loaded_module = importlib.import_module(module_name)
    except ImportError:
        raise exception_cls(f"Unable to load module: {module_name}")

    if not hasattr(loaded_module, name):
        raise exception_cls(
            f"Module ({module_name}) does not have a class called {name}"
        )

    return getattr(loaded_module, name)



"""Module containing a class to identify broken focussed bonds
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Sequence, Tuple
    from aizynthfinder.chem.mol import TreeMolecule
    from aizynthfinder.chem.reaction import RetroReaction


class BrokenBonds:
    """
    A class to keep track of focussed bonds breaking in a target molecule.

    :param focussed_bonds: A list of focussed bond pairs. The bond pairs are represented
        as tuples of size 2. These bond pairs exist in the target molecule's atom bonds.
    """

    def __init__(self, focussed_bonds: Sequence[Sequence[int]]) -> None:
        self.focussed_bonds = sort_bonds(focussed_bonds)
        self.filtered_focussed_bonds: List[Tuple[int, int]] = []

    def __call__(self, reaction: RetroReaction) -> List[Tuple[int, int]]:
        """
        Provides a list of focussed bonds that break in any of the molecule's reactants.

        :param reaction: A retro reaction.
        :return: A list of all the focussed bonds that broke within the reactants
            that constitute the target molecule.
        """
        self.filtered_focussed_bonds = self._get_filtered_focussed_bonds(reaction.mol)
        if not self.filtered_focussed_bonds:
            return []

        molecule_bonds = []
        for reactant in reaction.reactants[reaction.index]:
            molecule_bonds += reactant.mapped_atom_bonds

        broken_focussed_bonds = self._get_broken_frozen_bonds(
            sort_bonds(molecule_bonds)
        )
        return broken_focussed_bonds

    def _get_broken_frozen_bonds(
        self,
        molecule_bonds: List[Tuple[int, int]],
    ) -> List[Tuple[int, int]]:
        broken_focussed_bonds = list(
            set(self.filtered_focussed_bonds) - set(molecule_bonds)
        )
        return broken_focussed_bonds

    def _get_filtered_focussed_bonds(
        self, molecule: TreeMolecule
    ) -> List[Tuple[int, int]]:
        molecule_bonds = molecule.mapped_atom_bonds
        atom_maps = [atom_map for bonds in molecule_bonds for atom_map in bonds]

        filtered_focussed_bonds = []
        for idx1, idx2 in self.focussed_bonds:
            if idx1 in atom_maps and idx2 in atom_maps:
                filtered_focussed_bonds.append((idx1, idx2))
        return filtered_focussed_bonds


def sort_bonds(bonds: Sequence[Sequence[int]]) -> List[Tuple[int, int]]:
    return [tuple(sorted(bond)) for bond in bonds]  # type: ignore



""" Module containing routines to setup proper logging
"""
# pylint: disable=ungrouped-imports, wrong-import-order, wrong-import-position, unused-import
import logging.config
import os

import yaml

# See Github issue 30 why sklearn is imported here
try:
    import sklearn  # noqa
except ImportError:
    pass
from rdkit import RDLogger

from aizynthfinder.utils.paths import data_path
from aizynthfinder.utils.type_utils import Optional

# Suppress RDKit errors due to incomplete template (e.g. aromatic non-ring atoms)
rd_logger = RDLogger.logger()
rd_logger.setLevel(RDLogger.CRITICAL)


def logger() -> logging.Logger:
    """
    Returns the logger that should be used by all classes

    :return: the logger object
    """
    return logging.getLogger("aizynthfinder")


def setup_logger(
    console_level: int, file_level: Optional[int] = None
) -> logging.Logger:
    """
    Setup the logger that should be used by all classes

    The logger configuration is read from the `logging.yml` file.

    :param console_level: the level of logging to the console
    :param file_level: the level of logging to file, if not set logging to file is disabled, default to None
    :return: the logger object
    """
    filename = os.path.join(data_path(), "logging.yml")
    with open(filename, "r") as fileobj:
        config = yaml.load(fileobj.read(), Loader=yaml.SafeLoader)

    config["handlers"]["console"]["level"] = console_level
    if file_level:
        config["handlers"]["file"]["level"] = file_level
    else:
        del config["handlers"]["file"]
        config["loggers"]["aizynthfinder"]["handlers"].remove("file")

    logging.config.dictConfig(config)
    return logger()



""" Module containing classes that interfaces different stock classes
"""
from __future__ import annotations

import copy
from collections import defaultdict
from typing import TYPE_CHECKING

from aizynthfinder.chem import Molecule
from aizynthfinder.context.collection import ContextCollection
from aizynthfinder.context.stock.queries import (
    InMemoryInchiKeyQuery,
    MolbloomFilterQuery,
    STOCK_QUERY_ALIAS,
    StockQueryMixin,
)
from aizynthfinder.context.stock.queries import __name__ as queries_module
from aizynthfinder.utils.exceptions import StockException
from aizynthfinder.utils.loading import load_dynamic_class

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import (
        Any,
        Dict,
        List,
        Optional,
        Set,
        StrDict,
        Union,
    )


class Stock(ContextCollection):
    """
    A collection of molecules that are in stock

    A molecule can be queried on the stock with:

    .. code-block::

        my_mol = Molecule(smiles="CCO")
        my_mol in stock

    One can obtain individual stocks with:

    .. code-block::

        sub_stock = stock["key"]

    One can obtain the number of molecules in the selected stock with:

    .. code-block::

        number_of_molecules = len(stock)

    """

    _collection_name = "stock"

    def __init__(self) -> None:
        super().__init__()
        self._exclude: Set[str] = set()
        self._stop_criteria: StrDict = {"amount": None, "price": None, "counts": {}}
        self._use_stop_criteria: bool = False

    def __contains__(self, mol: Molecule) -> bool:
        if not self.selection or mol.inchi_key in self._exclude:
            return False

        if self._use_stop_criteria:
            return self._apply_stop_criteria(mol)

        for key in self.selection:
            if mol in self[key]:
                return True
        return False

    def __len__(self) -> int:
        return sum(len(self[key]) for key in self.selection or [])

    @property
    def stop_criteria(self) -> dict:
        """Return a copy of the stop criteria used by the stock"""
        return copy.deepcopy(self._stop_criteria)

    def amount(self, mol: Molecule) -> float:
        """
        Calculate the maximum amount of a molecule in stock

        :param mol: the molecule to query
        :raises StockException: if the amount could not be computed
        :return: the maximum amount
        """
        amounts = self._mol_property(mol, "amount")
        if not amounts:
            raise StockException("Could not obtain amount of molecule")
        return max(amounts)

    def availability_list(self, mol: Molecule) -> List[str]:
        """
        Return a list of what stocks a given mol is available

        If the molecule is not in stock it will return any empty list

        :param mol: The molecule to query
        :returns: string with a list of stocks that mol was found in
        """
        availability = []
        for key in self.selection or []:
            if mol not in self[key]:
                continue
            try:
                availability.extend(self[key].availability_string(mol).split(","))
            except (StockException, AttributeError):
                availability.append(key)
        return sorted(list(set(availability)))

    def availability_string(self, mol: Molecule) -> str:
        """
        Return a string of what stocks a given mol is available

        If the molecule is not in stock it will return "Not in stock"

        :param mol: The molecule to query
        :returns: string with a list of stocks that mol was found in
        """
        availability = self.availability_list(mol)
        if availability:
            return ",".join(availability)
        return "Not in stock"

    def exclude(self, mol: Molecule) -> None:
        """
        Exclude a molecule from the stock.
        When this molecule is queried it will return False,
        regardless if the molecule is in the stock.

        :param mol: the molecule to exclude
        """
        self._exclude.add(mol.inchi_key)

    def load(self, source: StockQueryMixin, key: str) -> None:  # type: ignore
        """
        Add a pre-initialized stock query object to the stock

        :param source: the item to add
        :param key: The key that will be used to select the stock
        """
        if not isinstance(source, StockQueryMixin):
            raise StockException(
                "Only objects of classes inherited from StockQueryMixin can be added"
            )

        self._logger.info(f"Loading stock from {source.__class__.__name__} to {key}")
        self._items[key] = source

    def load_from_config(self, **config: Any) -> None:
        """
        Load one or more stock queries from a configuration

        The key can be "stop_criteria" in case the config is given to the
        `set_stop_criteria` method

        The format should be
        key:
            type: name of the stock class or custom_package.custom_model.CustomClass
            path: path to the stock file
            other settings or params
        or
        key: path_to_model

        :param config: the configuration
        """
        if "stop_criteria" in config:
            self.set_stop_criteria(config["stop_criteria"])

        for key, stock_config in config.items():
            if key == "stop_criteria":
                continue

            if not isinstance(stock_config, dict):
                kwargs = {"path": stock_config}
                if stock_config.endswith(".bloom"):
                    cls: Any = MolbloomFilterQuery
                else:
                    cls = InMemoryInchiKeyQuery
            else:
                if "type" not in stock_config or stock_config["type"] == "inchiset":
                    cls = InMemoryInchiKeyQuery
                else:
                    stock_query = STOCK_QUERY_ALIAS.get(
                        stock_config["type"], stock_config["type"]
                    )
                    cls = load_dynamic_class(
                        stock_query, queries_module, StockException
                    )
                kwargs = dict(stock_config)

            if "type" in kwargs:
                del kwargs["type"]
            obj = cls(**kwargs)
            self.load(obj, key)

    def price(self, mol: Molecule) -> float:
        """
        Calculate the minimum price of a molecule in stock

        :param mol: the molecule to query
        :raises StockException: if the price could not be computed
        :return: the minimum price
        """
        prices = self._mol_property(mol, "price")
        if not prices:
            raise StockException("Could not obtain price of molecule")
        return min(prices)

    def reset_exclusion_list(self) -> None:
        """Remove all molecules in the exclusion list"""
        self._exclude = set()

    def select(self, value: Union[str, List[str]], append: bool = False) -> None:
        """
        Select one or more stock queries

        :param value: the key of the stocks to select
        :param append: if True and ``value`` is a single key append it to the current selection
        """
        super().select(value, append)
        try:
            self._logger.info(f"Compounds in stock: {len(self)}")
        except (TypeError, ValueError):  # In case len is not possible to compute
            pass

    def set_stop_criteria(self, criteria: Optional[Dict] = None) -> None:
        """
        Set criteria that stop the search

        The keys of the criteria can be "price" or "amount" which accepts numerical settings,
        or "counts" that should be dictionary of maximum allowed count for each atomic symbol.

        Example:

        .. code-block::

            criteria = {
                "price": 5,
                "amount": 100,
                "counts": {
                    "C": 6,
                    "O": 4
                }
            }

        :param criteria: the criteria settings
        """
        criteria = criteria or {}
        self._stop_criteria = {
            "price": criteria.get("price"),
            "amount": criteria.get("amount"),
            "counts": copy.deepcopy(criteria.get("size", criteria.get("counts"))),
        }
        self._use_stop_criteria = any(self._stop_criteria.values())
        reduced_criteria = {
            key: value for key, value in self._stop_criteria.items() if value
        }
        self._logger.info(f"Stop criteria for stock updated to: {reduced_criteria}")

    def smiles_in_stock(self, smiles: str) -> bool:
        """
        Check if the SMILES is in the currently selected stocks

        :param smiles: SMILES string (must be RDKit sanitizable)
        :returns: if the SMILES was on stock
        """
        return Molecule(smiles=smiles) in self

    def _apply_amount_criteria(self, mol: Molecule) -> bool:
        if not self._stop_criteria["amount"]:
            return True
        try:
            amount = self.amount(mol)
        except StockException:
            return True
        return amount >= self._stop_criteria.get("amount", amount)

    def _apply_counts_criteria(self, mol: Molecule) -> bool:
        if not self._stop_criteria["counts"]:
            return True
        atom_counts: dict = defaultdict(int)
        for atom in mol.rd_mol.GetAtoms():
            atom_counts[atom.GetSymbol()] += 1
        for symbol, threshold in self._stop_criteria["counts"].items():
            if atom_counts[symbol] > threshold:
                return False
        return True

    def _apply_price_criteria(self, mol: Molecule) -> bool:
        if not self._stop_criteria["price"]:
            return True
        try:
            price = self.price(mol)
        except StockException:
            return True
        return price <= self._stop_criteria.get("price", price)

    def _apply_stop_criteria(self, mol: Molecule) -> bool:
        if not self._apply_counts_criteria(mol):
            return False

        passes = False
        for key in self.selection or []:
            passes = passes or self[key].cached_search(mol)
        passes = passes and self._apply_amount_criteria(mol)
        passes = passes and self._apply_price_criteria(mol)

        for key in self.selection or []:
            self[key].clear_cache()
        return passes

    def _mol_property(self, mol, property_name):
        values = []
        for key in self.selection:
            try:
                func = getattr(self[key], property_name)
                values.append(func(mol))
            except StockException:
                pass
        return values



""" Module containing classes used to score the reaction routes.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from aizynthfinder.context.collection import ContextCollection
from aizynthfinder.context.scoring.scorers import (
    AverageTemplateOccurrenceScorer,
    NumberOfPrecursorsInStockScorer,
    NumberOfPrecursorsScorer,
    NumberOfReactionsScorer,
    Scorer,
    StateScorer,
)
from aizynthfinder.context.scoring.scorers import __name__ as scorers_module
from aizynthfinder.reactiontree import ReactionTree
from aizynthfinder.search.mcts import MctsNode
from aizynthfinder.utils.exceptions import ScorerException
from aizynthfinder.utils.loading import load_dynamic_class

if TYPE_CHECKING:
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import Any, List, Sequence, TypeVar, Union

    _Scoreable = TypeVar("_Scoreable", MctsNode, ReactionTree)
    _Scoreables = Sequence[_Scoreable]
    _ScorerItemType = Union[_Scoreables, _Scoreable]


_SIMPLE_SCORERS = [
    StateScorer,
    NumberOfReactionsScorer,
    NumberOfPrecursorsScorer,
    NumberOfPrecursorsInStockScorer,
    AverageTemplateOccurrenceScorer,
]


class ScorerCollection(ContextCollection):
    """
    Store scorer objects for the aizynthfinder interface.

    The scorers can be obtained by name with simple indexing

    .. code-block::

        scorers = ScorerCollection()
        scorer = scorers['state score']

    Scorers defined in this module and that does not require any
    other argument to initialize than the ``config`` are auto-loaded.

    :param config: the configuration of the tree search
    """

    _collection_name = "scorer"

    def __init__(self, config: Configuration) -> None:
        super().__init__()
        self._config = config
        self.create_default_scorers()

    def __repr__(self) -> str:
        if self.selection:
            return f"{self._collection_name} ({', '.join(self.selection)})"

        return f"{self._collection_name} ({', '.join(self.items)})"

    def create_default_scorers(self) -> None:
        """
        Setup the scores that only need the config as their input.
        """
        for cls in _SIMPLE_SCORERS:
            self.load(cls(self._config), silent=True)

    def load(self, scorer: Scorer, silent: bool = False) -> None:  # type: ignore
        """
        Add a pre-initialized scorer object to the collection

        :param scorer: the item to add
        :param silent: if True will not write out the name of the loaded scorer
        """
        if not isinstance(scorer, Scorer):
            raise ScorerException(
                "Only objects of classes inherited from Scorer can be added"
            )
        self._items[repr(scorer)] = scorer
        if not silent:
            self._logger.info(f"Loaded scorer: {repr(scorer)}")

    def load_from_config(self, **scorers_config: Any) -> None:
        """
        Load one or several scorers from a configuration dictionary

        The keys are the name of scorer class. If a scorer is not
        defined in the ``aizynthfinder.context.scoring`` module, the module
        name can be appended, e.g. ``mypackage.scoring.AwesomeScore``.

        The values of the configuration is passed directly to the scorer
        class along with the ``config`` parameter.

        :raises ScorerException: if module or class could not be found
        """
        for name_spec, scorer_config in scorers_config.items():
            config_str = (
                f" from configuration '{scorer_config}'" if scorer_config else ""
            )
            cls = load_dynamic_class(name_spec, scorers_module, ScorerException)
            obj = cls(self._config, **(scorer_config or {}))
            self._logger.info(f"Loaded scorer: '{repr(obj)}'{config_str}")
            self._items[repr(obj)] = obj

    def make_subset(self, subset_names: List[str]) -> "ScorerCollection":
        """
        Make a new scorer collection by taking a subset of this
        collection. The scorer instances will be shared between
        the collections

        :param subset_names: the scorers to copy over
        :returns: the newly formed collection
        """
        new_collection = ScorerCollection(self._config)
        for name in new_collection.names():
            del new_collection[name]
        for name in subset_names:
            try:
                scorer = self._items[name]
            except KeyError:
                raise ScorerException(f"Unable to find '{name}' in parent collection")
            new_collection.load(scorer, silent=True)
            new_collection.select(name, append=True)
        return new_collection

    def names(self) -> List[str]:
        """Return a list of the names of all the loaded scorers"""
        return self.items

    def objects(self) -> List[Scorer]:
        """Return a list of all the loaded scorer objects"""
        return list(self._items.values())

    def score_vector(self, item: _Scoreable) -> Sequence[float]:
        """
        For the given item, score it with all selected scorers
        and return a vector

        :param item: the item to be scored
        :returns: the vector with the scores
        """
        if not self.selection:
            return []
        return [self[scorer](item) for scorer in self.selection]

    def weighted_score(self, item: _Scoreable, weights: Sequence[float]) -> float:
        """
        For the given item, score it with all selected scorers
        and return a weighted sum of all the scores.

        If weights is not the same length as the number of scorers
        an exception is raised.

        If no scorers are selected this will raise an exception

        :param item: the item to be scored
        :param weights: the weights of the scorers
        :returns: the weighted sum
        """
        if not self.selection:
            raise ScorerException(
                "No scorers are selected so cannot compute weighted sum"
            )

        if len(weights) != len(self.selection):
            raise ScorerException(
                "The number of weights given does not agree with the number of scorers"
            )
        return sum(
            weight * score for weight, score in zip(weights, self.score_vector(item))
        )



""" Module containing a class that represents a node in the search tree.
"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING

import numpy as np
from paretoset import paretoset

from aizynthfinder.chem import TreeMolecule, deserialize_action, serialize_action
from aizynthfinder.search.mcts.state import MctsState
from aizynthfinder.search.mcts.utils import ReactionTreeFromSuperNode, route_to_node
from aizynthfinder.utils.exceptions import (
    NodeUnexpectedBehaviourException,
    RejectionException,
)
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.chem import (
        MoleculeDeserializer,
        MoleculeSerializer,
        RetroReaction,
    )
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.reactiontree import ReactionTree
    from aizynthfinder.search.mcts.search import MctsSearchTree
    from aizynthfinder.utils.type_utils import List, Optional, StrDict, Tuple


class MctsNode:
    """
    A node in the search tree.

    The children are instantiated lazily for efficiency: only when
    a child is selected the reaction to create that child is applied.

    Properties of an instantiated children to a node can be access with:

    .. code-block::

        children_attr = node[child]

    the return value is a dictionary with keys "action", "value", "prior"
    and "visitations".

    :ivar is_expanded: if the node has had children added to it
    :ivar is_expandable: if the node is expandable
    :ivar tree: the tree owning this node

    :param state: the state of the node
    :param owner: the tree that owns this node
    :param config: settings of the tree search algorithm
    :param parent: the parent node, defaults to None
    """

    def __init__(
        self,
        state: MctsState,
        owner: MctsSearchTree,
        config: Configuration,
        parent: Optional[MctsNode] = None,
    ):
        self._state = state
        self._config = config
        self._expansion_policy = config.expansion_policy
        self._filter_policy = config.filter_policy
        self.tree = owner
        self.is_expanded: bool = False
        self.is_expandable: bool = not self.state.is_terminal
        self._parent = parent

        if owner is None:
            self.created_at_iteration: Optional[int] = None
        else:
            self.created_at_iteration = self.tree.profiling["iterations"]

        self._children_values: List[float] = []
        self._children_priors: List[float] = []
        self._children_visitations: List[int] = []
        self._children_actions: List[RetroReaction] = []
        self._children: List[Optional[MctsNode]] = []

        self.blacklist = set(mol.inchi_key for mol in state.expandable_mols)
        if parent:
            self.blacklist = self.blacklist.union(parent.blacklist)

        if self._algo_config["mcts_grouping"]:
            self._degeneracy_check = self._algo_config["mcts_grouping"].lower()
        else:
            self._degeneracy_check = "none"
        self._logger = logger()

    def __getitem__(self, node: "MctsNode") -> StrDict:
        idx = self._children.index(node)
        return {
            "action": self._children_actions[idx],
            "value": self._children_values[idx],
            "prior": self._children_priors[idx],
            "visitations": self._children_visitations[idx],
        }

    @classmethod
    def create_root(
        cls, smiles: str, tree: MctsSearchTree, config: Configuration
    ) -> "MctsNode":
        """
        Create a root node for a tree using a SMILES.

        :param smiles: the SMILES representation of the root state
        :param tree: the search tree
        :param config: settings of the tree search algorithm
        :return: the created node
        """
        mol = TreeMolecule(parent=None, transform=0, smiles=smiles)
        state = MctsState(mols=[mol], config=config)
        return cls(state=state, owner=tree, config=config)

    @classmethod
    def from_dict(
        cls,
        dict_: StrDict,
        tree: MctsSearchTree,
        config: Configuration,
        molecules: MoleculeDeserializer,
        parent: Optional["MctsNode"] = None,
    ) -> "MctsNode":
        """
        Create a new node from a dictionary, i.e. deserialization.

        :param dict_: the serialized node
        :param tree: the search tree
        :param config: settings of the tree search algorithm
        :param molecules: the deserialized molecules
        :param parent: the parent node
        :return: a deserialized node
        """
        # pylint: disable=protected-access
        state = MctsState.from_dict(dict_["state"], config, molecules)
        node = cls(state=state, owner=tree, config=config, parent=parent)
        node.is_expanded = dict_["is_expanded"]
        node.is_expandable = dict_["is_expandable"]
        node._children_values = dict_["children_values"]
        node._children_priors = dict_["children_priors"]
        node._children_visitations = dict_["children_visitations"]
        node._children_actions = [
            deserialize_action(action_dict, molecules)
            for action_dict in dict_["children_actions"]
        ]
        node._children = [
            cls.from_dict(child, tree, config, molecules, parent=node)
            if child
            else None
            for child in dict_["children"]
        ]
        return node

    @property
    def children(self) -> List["MctsNode"]:
        """
        Returns all of the instantiated children.

        :return: the children
        """
        return [child for child in self._children if child]

    @property
    def is_solved(self) -> bool:
        """Return if the state is solved."""
        return self.state.is_solved

    @property
    def parent(self) -> Optional["MctsNode"]:
        """Return the parent of the node."""
        return self._parent

    @property
    def state(self) -> MctsState:
        """Return the underlying state of the node."""
        return self._state

    @property
    def _algo_config(self) -> StrDict:
        """Just a convinient, shorter name of this."""
        return self._config.search.algorithm_config

    def actions_to(self) -> List[RetroReaction]:
        """
        Returns the actions leading to this node

        :return: the list of actions
        """
        return self.path_to()[0]

    def backpropagate(self, child: "MctsNode", value_estimate: float) -> None:
        """
        Update the number of visitations of a particular child and its value.

        :param child: the child node
        :param value_estimate: the value to add to the child value
        """
        idx = self._children.index(child)
        self._children_visitations[idx] += 1
        self._children_values[idx] += value_estimate

    def children_view(self) -> StrDict:
        """
        Creates a view of the children attributes. Each of the
        list returned is a new list, although the actual children
        are not copied.

        The return dictionary will have keys "actions", "values",
        "priors", "visitations" and "objects".

        :return: the view
        """
        return {
            "actions": list(self._children_actions),
            "values": list(self._children_values),
            "priors": list(self._children_priors),
            "visitations": list(self._children_visitations),
            "objects": list(self._children),
        }

    def expand(self) -> None:
        """
        Expand the node.

        Expansion is the process of creating the children of the node,
        without instantiating a child object. The actions and priors are
        taken from the policy network.

        If immediate instantiation is marked for some policies, however, the
        children nodes will be instantiated.
        """
        if self.is_expanded:
            msg = f"Oh no! This node is already expanded. id={id(self)}"
            self._logger.debug(msg)
            raise NodeUnexpectedBehaviourException(msg)

        if self.is_expanded or not self.is_expandable:
            return

        self.is_expanded = True

        cache_molecules = []
        if self.parent:
            for child in self.parent.children:
                if child is not self:
                    cache_molecules.extend(child.state.expandable_mols)

        # Calculate the possible actions, fill the child_info lists
        # Actions by default only assumes 1 set of reactants
        actions, priors = self._expansion_policy(
            self.state.expandable_mols, cache_molecules
        )
        self._fill_children_lists(actions, priors)

        # Reverse the expansion if it did not produce any children
        if len(actions) == 0:
            self.is_expandable = False
            self.is_expanded = False

        if self.tree:
            self.tree.profiling["expansion_calls"] += 1

        if not self._algo_config["immediate_instantiation"]:
            return
        # Instantiate all children actions created by the marked policy,
        # a new list of actions will be iterated over, because it can grow due
        # to instantiation
        nactions = len(actions)
        for child_idx, action in enumerate(self._children_actions[:nactions]):
            policy_name = action.metadata.get("policy_name")
            if (
                policy_name
                and policy_name in self._algo_config["immediate_instantiation"]
            ):
                self._instantiate_child(child_idx)

    def is_terminal(self) -> bool:
        """
        Node is terminal if its unexpandable, or the internal state is terminal (solved).

        :return: the terminal attribute of the node
        """
        return not self.is_expandable or self.state.is_terminal

    def path_to(self) -> Tuple[List[RetroReaction], List[MctsNode]]:
        """
        Return the path to this node, which is a list of actions and a list of node.

        :return: the actions and nodes
        """
        return route_to_node(self)

    def promising_child(self) -> Optional["MctsNode"]:
        """
        Return the child with the currently highest Q+U.

        The selected child will be instantiated if it has not been already.

        If no actions could be found that were applicable, the method will
        return None.

        :return: the child
        """
        child = None
        while child is None:
            try:
                child = self._score_and_select()
            # _score_and_select raises exception if no children can be selected
            except ValueError:
                child = None
                break

        if not child:
            self._logger.debug(
                "Returning None from promising_child() because there were no applicable action"
            )
            self.is_expanded = False
            self.is_expandable = False

        return child

    def serialize(self, molecule_store: MoleculeSerializer) -> StrDict:
        """
        Serialize the node object to a dictionary.

        :param molecule_store: the serialized molecules
        :return: the serialized node
        """
        return {
            "state": self.state.serialize(molecule_store),
            "children_values": self._serialize_stats_list("_children_values"),
            "children_priors": self._serialize_stats_list("_children_priors"),
            "children_visitations": self._children_visitations,
            "children_actions": [
                serialize_action(action, molecule_store)
                for action in self._children_actions
            ],
            "children": [
                child.serialize(molecule_store) if child else None
                for child in self._children
            ],
            "is_expanded": self.is_expanded,
            "is_expandable": self.is_expandable,
        }

    def to_reaction_tree(self) -> ReactionTree:
        """
        Return reaction tree from the path of actions and nodes leading to this node.

        :return: the constructed tree
        """
        return ReactionTreeFromSuperNode(self).tree

    def _check_child_reaction(self, reaction: RetroReaction) -> bool:
        if not reaction.reactants:
            self._logger.debug(f"{reaction} did not produce any reactants")
            return False

        # fmt: off
        reactants0 = reaction.reactants[0]
        if len(reaction.reactants) == 1 and len(reactants0) == 1 and reaction.mol == reactants0[0]:
            return False
        # fmt: on

        return True

    def _children_q(self) -> np.ndarray:
        return np.array(self._children_values) / np.array(self._children_visitations)

    def _children_u(self) -> np.ndarray:
        total_visits = np.log(np.sum(self._children_visitations))
        child_visits = np.array(self._children_visitations)
        return self._algo_config["C"] * np.sqrt(2 * total_visits / child_visits)

    def _create_children_nodes(
        self, states: List[MctsState], child_idx: int
    ) -> List["MctsNode"]:
        new_nodes = []
        first_child_idx = child_idx
        for state_index, state in enumerate(states):
            if self._generated_degeneracy(state, first_child_idx):
                # Only need to disable first new child,
                # if the action generated more states, we will just not generate
                # a child for that state
                if state_index == 0:
                    self._disable_child(child_idx)
                continue

            # If there's more than one outcome, the lists need be expanded
            if state_index > 0:
                child_idx = self._expand_children_lists(first_child_idx, state_index)

            if self._filter_child_reaction(self._children_actions[child_idx]):
                self._disable_child(child_idx)
            else:
                new_node = self.__class__(
                    state=state, owner=self.tree, config=self._config, parent=self
                )
                self._children[child_idx] = new_node
                new_nodes.append(new_node)
        return new_nodes

    def _disable_child(self, child_idx: int) -> None:
        self._children_values[child_idx] = -1e6

    def _expand_children_lists(self, old_index: int, action_index: int) -> int:
        new_action = self._children_actions[old_index].copy(index=action_index)
        self._children_actions.append(new_action)
        self._children_priors.append(self._children_priors[old_index])
        self._children_values.append(self._children_values[old_index])
        self._children_visitations.append(self._children_visitations[old_index])
        self._children.append(None)
        return len(self._children) - 1

    def _fill_children_lists(
        self, actions: List[RetroReaction], priors: List[float]
    ) -> None:
        self._children_actions = actions
        self._children_priors = priors
        nactions = len(actions)
        self._children_visitations = [1] * nactions
        self._children = [None] * nactions
        if self._algo_config["use_prior"]:
            self._children_values = list(self._children_priors)
        else:
            self._children_values = [self._algo_config["default_prior"]] * nactions

    def _filter_child_reaction(self, reaction: RetroReaction) -> bool:
        if self._regenerated_blacklisted(reaction):
            self._logger.debug(
                f"Reaction {reaction.reaction_smiles()} "
                f"was rejected because it re-generated molecule not in stock"
            )
            return True

        if not self._filter_policy.selection:
            return False
        try:
            self._filter_policy(reaction)
        except RejectionException as err:
            self._logger.debug(str(err))
            return True
        return False

    def _generated_degeneracy(self, new_state: MctsState, child_idx: int) -> bool:
        """
        Check if a new MCTS state is equal to another MCTS state of a children node.

        The check can be "partial" in which the equality is based only on the expandable molecules,
        or "full" in which the equality is based on all molecules in the state.

        The comparison will not be made on unexpanded children nodes
        or terminal children nodes.

        The metadata of the degenerate action will be added to the metadata
        of the previously created equal state.
        """

        def equal_states(query_state):
            if self._degeneracy_check == "partial":
                return query_state.expandables_hash == new_state.expandables_hash
            return query_state == new_state

        if self._degeneracy_check not in ["partial", "full"]:
            return False
        previous_action = None
        for child, action in zip(self._children, self._children_actions):
            if (
                child is not None
                and not child.is_terminal()
                and equal_states(child.state)
            ):
                previous_action = action
                break

        if previous_action is None:
            return False

        # No need to copy the metadata because it will be the same
        if previous_action is self._children_actions[child_idx]:
            return True

        metadata_copy = dict(self._children_actions[child_idx].metadata)
        if "additional_actions" not in previous_action.metadata:
            previous_action.metadata["additional_actions"] = []
        previous_action.metadata["additional_actions"].append(metadata_copy)
        return True

    def _instantiate_child(self, child_idx: int) -> List["MctsNode"]:
        """
        Instantiate the children node.

        The algorithm is:
        * Apply the reaction associated with the child
        * If the application of the action failed, set value to -1e6 and return None
        * Create a new state array, one new state for each of the reaction outcomes
        * Create new child nodes
            - If a filter policy is available and the reaction outcome is unlikely
              set value of child to -1e6
         * Return all new nodes
        """
        if self._children[child_idx] is not None:
            raise NodeUnexpectedBehaviourException("Node already instantiated")

        reaction = self._children_actions[child_idx]
        if reaction.unqueried:
            if self.tree:
                self.tree.profiling["reactants_generations"] += 1
            _ = reaction.reactants

        if not self._check_child_reaction(reaction):
            self._disable_child(child_idx)
            return []

        keep_mols = [mol for mol in self.state.mols if mol is not reaction.mol]
        new_states = [
            MctsState(keep_mols + list(reactants), self._config)
            for reactants in reaction.reactants
        ]
        return self._create_children_nodes(new_states, child_idx)

    def _regenerated_blacklisted(self, reaction: RetroReaction) -> bool:
        if not self._algo_config["prune_cycles_in_search"]:
            return False
        for reactants in reaction.reactants:
            for mol in reactants:
                if mol.inchi_key in self.blacklist:
                    return True
        return False

    def _score_and_select(self) -> Optional["MctsNode"]:
        if not max(self._children_values) > 0:
            raise ValueError("Has no selectable children")
        scores = self._children_q() + self._children_u()
        indices = np.where(scores == scores.max())[0]
        index = np.random.choice(indices)
        return self._select_child(index)

    def _select_child(self, child_idx: int) -> Optional["MctsNode"]:
        """
        Selecting a child node implies instantiating the children nodes.

        If the child has already been instantiated, return immediately
        Otherwise, select a random node of the feasible ones to return
        """
        if self._children[child_idx]:
            return self._children[child_idx]

        new_nodes = self._instantiate_child(child_idx)
        if new_nodes:
            return random.choice(new_nodes)
        return None

    def _serialize_stats_list(self, name: str) -> List[float]:
        return [float(value) for value in getattr(self, name)]


class ParetoMctsNode(MctsNode):
    """
    A node in a multi-objective tree search.

    This implements the algorithm from:
        Chen W., Liu L. Pareto Monte Carlo Tree Search for Multi-Objective Informative Planning
        Robotics: Science and Systems 2019, 2012 arXiv:2111.01825


    The main difference compared to the standard MCTS algorithm is:
        - Children stats: the values, cumulative reward and priors are nested
                     list, with one value per objective
        - Selection: children on Pareto front are computed, and
                     a child from this set is taken randomly

    This implementation disregards the prior of the children when it has been
    visited once.

    It is assumed that all objectives are to be maximised.
    """

    def __init__(
        self,
        state: MctsState,
        owner: MctsSearchTree,
        config: Configuration,
        parent: Optional[ParetoMctsNode] = None,
    ):
        super().__init__(state, owner, config, parent)
        self._num_objectives = len(self._algo_config["search_rewards"])
        self._prior_weight = 1
        self._direction = "max"  # current implementation assumes maximisation
        self._children_rewards_cummulative: List[List[float]]
        self._children_values: List[List[float]]  # type: ignore
        self._children_priors: List[List[float]]  # type: ignore

    def backpropagate(self, child: "MctsNode", value_estimate: List[float]) -> None:  # type: ignore
        """
        Update the number of visitations of a particular child and its value.

        :param child: the child node
        :param value_estimate: the value to add to the child value
        """
        idx = self._children.index(child)
        self._children_visitations[idx] += 1
        # here we only update the cummulative rewards,
        #  _children_values are updated at selection time
        new_value_estimate = [
            cum_reward + new_reward
            for cum_reward, new_reward in zip(
                self._children_rewards_cummulative[idx], value_estimate
            )
        ]
        self._children_rewards_cummulative[idx] = new_value_estimate

    def children_view(self) -> StrDict:
        """
        Creates a view of the children attributes. Each of the
        list returned is a new list, although the actual children
        are not copied.

        The return dictionary will have keys "actions", "values",
        "priors", "visitations", "rewards_cum", and "objects".

        :return: the view
        """
        dict_ = super().children_view()
        dict_["rewards_cum"] = list(self._children_rewards_cummulative)
        return dict_

    def serialize(self, molecule_store: MoleculeSerializer) -> StrDict:
        """
        Serialize the node object to a dictionary.

        :param molecule_store: the serialized molecules
        :return: the serialized node
        """
        dict_ = super().serialize(molecule_store)
        dict_["children_cumulative_reward"] = self._serialize_stats_list(
            "_children_rewards_cummulative"
        )
        return dict_

    def _disable_child(self, child_idx: int) -> None:
        self._children_rewards_cummulative[child_idx] = [-1e6] * self._num_objectives

    def _expand_children_lists(self, old_index: int, action_index: int) -> int:
        ret = super()._expand_children_lists(old_index, action_index)
        # These lists are nested lists, so need to make sure that the inner lists are also new
        self._children_values[-1] = list(self._children_values[-1])
        self._children_priors[-1] = list(self._children_priors[-1])
        self._children_rewards_cummulative.append(
            list(self._children_rewards_cummulative[old_index])
        )
        return ret

    def _fill_children_lists(
        self, actions: List[RetroReaction], priors: List[float]
    ) -> None:
        self._children_actions = actions
        nactions = len(actions)
        # shape: num_actions x 1
        self._children_visitations = [1] * nactions
        self._children = [None] * nactions
        # shape: num_actions x num_objectives
        self._children_rewards_cummulative = [[0.0] * self._num_objectives] * nactions
        if self._algo_config["use_prior"]:
            # shape: num_actions x num_objectives
            # for children i, 3 objectives -> [prior i, prior i, prior i]
            self._children_priors = [[prior] * self._num_objectives for prior in priors]

        else:
            self._children_priors = [
                [self._algo_config["default_prior"]] * self._num_objectives
            ] * nactions

        # at initialisation, values = prior as cummulative rewards are zero
        self._children_values = [
            [prior * self._prior_weight for prior in priors]
            for priors in self._children_priors
        ]

    def _children_q(self, children_values_arr):
        children_visitations_expanded = np.repeat(
            np.array(self._children_visitations).reshape(-1, 1),
            axis=1,
            repeats=self._num_objectives,
        )
        return children_values_arr / children_visitations_expanded

    def _compute_children_scores(self) -> np.ndarray:
        """Compute the modified ucb scores: alpha * prior + average reward + exploration."""
        # update prior to zero once the node has been visited
        children_priors_arr = self._prior_schedule_oneoff()
        # compute prior_weight * prior + cummulative rewards
        children_values_arr = self._prior_weight * children_priors_arr + np.array(
            self._children_rewards_cummulative
        )
        expanded_u = np.repeat(
            self._children_u().reshape(-1, 1), axis=1, repeats=self._num_objectives
        )
        # _children_scores shape: num_childrens x num_objectives
        children_scores = self._children_q(children_values_arr) + expanded_u
        if children_scores.shape[1] != self._num_objectives:
            raise ValueError(
                f"expected second dimension to have {self._num_objectives},"
                f"currently has {children_scores.shape[1]}"
            )
        self._children_values = children_values_arr.tolist()
        self._children_priors = children_priors_arr.tolist()
        return children_scores

    def _prior_schedule_oneoff(self) -> np.ndarray:
        # shape: num_children x 1
        visted_mask = (np.array(self._children_visitations) > 1).reshape(-1, 1)
        # shape: num_children x num_objectives
        visted_mask = np.repeat(visted_mask, axis=1, repeats=self._num_objectives)
        # set the prior weights for visited children to be zero
        children_priors_arr = np.array(self._children_priors)
        children_priors_arr[visted_mask] = 0
        return children_priors_arr

    def _score_and_select(self) -> Optional["MctsNode"]:
        if not max(max(value_list) for value_list in self._children_values) > 0:
            raise ValueError("Has no selectable children")
        children_scores = self._compute_children_scores()
        pareto_idxs = self._update_pareto_front(children_scores)
        index = np.random.choice(pareto_idxs)
        return self._select_child(index)

    def _serialize_stats_list(self, name: str):
        return [
            [float(value) for value in value_list] for value_list in getattr(self, name)
        ]

    def _update_pareto_front(self, children_scores: np.ndarray) -> np.ndarray:
        """
        Update the pareto front of a node, this step normally happens
        after its values for the best child have been updated.

        :param children_scores: Children scores
        :returns: Pareto front children indexes
        """
        direction_arr = np.repeat(self._direction, self._num_objectives)
        mask = paretoset(children_scores, sense=direction_arr, distinct=False)
        return np.arange(len(self._children))[mask]



""" Module containing classes that interfaces neural network policies
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from aizynthfinder.context.collection import ContextCollection
from aizynthfinder.context.policy.expansion_strategies import (
    ExpansionStrategy,
    TemplateBasedExpansionStrategy,
)
from aizynthfinder.context.policy.expansion_strategies import (
    __name__ as expansion_strategy_module,
)
from aizynthfinder.context.policy.filter_strategies import (
    FILTER_STRATEGY_ALIAS,
    FilterStrategy,
    QuickKerasFilter,
)
from aizynthfinder.context.policy.filter_strategies import (
    __name__ as filter_strategy_module,
)
from aizynthfinder.utils.exceptions import PolicyException
from aizynthfinder.utils.loading import load_dynamic_class

if TYPE_CHECKING:
    from aizynthfinder.chem import TreeMolecule
    from aizynthfinder.chem.reaction import RetroReaction
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import Any, Dict, List, Sequence, Tuple


class ExpansionPolicy(ContextCollection):
    """
    An abstraction of an expansion policy.

    This policy provides actions that can be applied to a molecule

    :param config: the configuration of the tree search
    """

    _collection_name = "expansion policy"

    def __init__(self, config: Configuration) -> None:
        super().__init__()
        self._config = config

    def __call__(
        self,
        molecules: Sequence[TreeMolecule],
        cache_molecules: Sequence[TreeMolecule] = None,
    ) -> Tuple[List[RetroReaction], List[float]]:
        return self.get_actions(molecules, cache_molecules)

    def get_actions(
        self,
        molecules: Sequence[TreeMolecule],
        cache_molecules: Sequence[TreeMolecule] = None,
    ) -> Tuple[List[RetroReaction], List[float]]:
        """
        Get all the probable actions of a set of molecules, using the selected policies

        :param molecules: the molecules to consider
        :param cache_molecules: additional molecules that potentially are sent to
                                  the expansion model but for which predictions are not returned
        :return: the actions and the priors of those actions
        :raises: PolicyException: if the policy isn't selected
        """
        if not self.selection:
            raise PolicyException("No expansion policy selected")

        all_possible_actions = []
        all_priors = []
        for name in self.selection:
            possible_actions, priors = self[name].get_actions(
                molecules, cache_molecules
            )
            all_possible_actions.extend(possible_actions)
            all_priors.extend(priors)
        return all_possible_actions, all_priors

    def load(self, source: ExpansionStrategy) -> None:  # type: ignore
        """
        Add a pre-initialized expansion strategy object to the policy

        :param source: the item to add
        """
        if not isinstance(source, ExpansionStrategy):
            raise PolicyException(
                "Only objects of classes inherited from ExpansionStrategy can be added"
            )
        self._items[source.key] = source

    def load_from_config(self, **config: Any) -> None:
        """
        Load one or more expansion policy from a configuration

        The format should be
        key:
            type: name of the expansion class or custom_package.custom_model.CustomClass
            model: path_to_model
            template: path_to_templates
            other settings or params
        or
        key:
            - path_to_model
            - path_to_templates

        :param config: the configuration
        """
        for key, strategy_config in config.items():
            if not isinstance(strategy_config, dict):
                model, template = strategy_config
                kwargs = {"model": model, "template": template}
                cls = TemplateBasedExpansionStrategy
            else:
                if (
                    "type" not in strategy_config
                    or strategy_config["type"] == "template-based"
                ):
                    cls = TemplateBasedExpansionStrategy
                else:
                    cls = load_dynamic_class(
                        strategy_config["type"],
                        expansion_strategy_module,
                        PolicyException,
                    )
                kwargs = dict(strategy_config)

            if "type" in kwargs:
                del kwargs["type"]
            obj = cls(key, self._config, **kwargs)
            self.load(obj)

    def reset_cache(self) -> None:
        """
        Reset the cache on all loaded policies
        """
        for policy in self._items.values():
            policy.reset_cache()


class FilterPolicy(ContextCollection):
    """
    An abstraction of a filter policy.

    This policy provides a query on a reaction to determine whether it should be rejected

    :param config: the configuration of the tree search
    """

    _collection_name = "filter policy"

    def __init__(self, config: Configuration) -> None:
        super().__init__()
        self._config = config

    def __call__(self, reaction: RetroReaction) -> None:
        return self.apply(reaction)

    def apply(self, reaction: RetroReaction) -> None:
        """
        Apply the all the selected filters on the reaction. If the reaction
        should be rejected a `RejectionException` is raised

        :param reaction: the reaction to filter
        :raises: if the reaction should be rejected or if a policy is selected
        """
        if not self.selection:
            raise PolicyException("No filter policy selected")

        for name in self.selection:
            self[name](reaction)

    def load(self, source: FilterStrategy) -> None:  # type: ignore
        """
        Add a pre-initialized filter strategy object to the policy

        :param source: the item to add
        """
        if not isinstance(source, FilterStrategy):
            raise PolicyException(
                "Only objects of classes inherited from FilterStrategy can be added"
            )
        self._items[source.key] = source

    def load_from_config(self, **config: Any) -> None:
        """
        Load one or more filter policy from a configuration

        The format should be
        key:
            type: name of the filter class or custom_package.custom_model.CustomClass
            model: path_to_model
            other settings or params
        or
        key: path_to_model

        :param config: the configuration
        """
        for key, strategy_config in config.items():
            if not isinstance(strategy_config, dict):
                model = strategy_config
                kwargs = {"model": model}
                cls = QuickKerasFilter
            else:
                if (
                    "type" not in strategy_config
                    or strategy_config["type"] == "quick-filter"
                ):
                    cls = QuickKerasFilter
                else:
                    strategy_spec = FILTER_STRATEGY_ALIAS.get(
                        strategy_config["type"], strategy_config["type"]
                    )
                    cls = load_dynamic_class(
                        strategy_spec, filter_strategy_module, PolicyException
                    )
                kwargs = dict(strategy_config)

            if "type" in kwargs:
                del kwargs["type"]
            obj = cls(key, self._config, **kwargs)
            self.load(obj)

    def reset_cache(self) -> None:
        """Reset filtering cache."""
        if not self.selection:
            return

        for name in self.selection:
            if hasattr(self[name], "reset_cache"):
                self[name].reset_cache()



""" Module containing classes and routines for the CLI
"""

from __future__ import annotations

import argparse
import importlib
import json
import logging
import os
import tempfile
import time
from collections import defaultdict
from typing import TYPE_CHECKING

import pandas as pd

from aizynthfinder.aizynthfinder import AiZynthFinder
from aizynthfinder.chem import Molecule
from aizynthfinder.utils.files import (
    cat_datafiles,
    save_datafile,
    split_file,
    start_processes,
)
from aizynthfinder.utils.logging import logger, setup_logger

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import (
        Any,
        Callable,
        Dict,
        List,
        Optional,
        StrDict,
        Union,
    )

    _PostProcessingJob = Callable[[AiZynthFinder], StrDict]
    _PreProcessingJob = Callable[[AiZynthFinder, int], None]


def _do_clustering(
    finder: AiZynthFinder,
    results: StrDict,
    detailed_results: bool,
    model_path: Optional[str] = None,
) -> None:
    time0 = time.perf_counter_ns()
    if model_path:
        kwargs = {"distances_model": "lstm", "model_path": model_path}
    else:
        kwargs = {"distances_model": "ted"}
    results["cluster_labels"] = finder.routes.cluster(n_clusters=0, **kwargs)  # type: ignore
    if not detailed_results:
        return

    results["cluster_time"] = (time.perf_counter_ns() - time0) * 1e-9
    results["distance_matrix"] = finder.routes.distance_matrix().tolist()


def _do_post_processing(
    finder: AiZynthFinder, results: StrDict, jobs: List[_PostProcessingJob]
) -> None:
    for job in jobs:
        results.update(job(finder))


def _get_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser("aizynthcli")
    parser.add_argument(
        "--smiles",
        required=True,
        help="the target molecule smiles or the path of a file containing the smiles",
    )
    parser.add_argument(
        "--config", required=True, help="the filename of a configuration file"
    )
    parser.add_argument(
        "--policy",
        nargs="+",
        default=[],
        help="the name of the expansion policy to use",
    )
    parser.add_argument(
        "--filter", nargs="+", default=[], help="the name of the filter to use"
    )
    parser.add_argument(
        "--stocks", nargs="+", default=[], help="the name of the stocks to use"
    )
    parser.add_argument(
        "--output", help="the name of the output file (JSON or HDF5 file)"
    )
    parser.add_argument(
        "--log_to_file",
        action="store_true",
        default=False,
        help="if provided, detailed logging to file is enabled",
    )
    parser.add_argument(
        "--nproc",
        type=int,
        help="if given, the input is split over a number of processes",
    )
    parser.add_argument(
        "--cluster",
        action="store_true",
        default=False,
        help="if provided, perform automatic clustering",
    )
    parser.add_argument(
        "--route_distance_model",
        help="if provided, calculate route distances for clustering with this ML model",
    )
    parser.add_argument(
        "--post_processing",
        nargs="+",
        help="a number of modules that performs post-processing tasks",
    )
    parser.add_argument(
        "--pre_processing", help="a module that perform pre-processing tasks"
    )
    parser.add_argument(
        "--checkpoint",
        required=False,
        help="the path to the checkpoint file",
    )
    return parser.parse_args()


def _load_postprocessing_jobs(modules: Optional[List[str]]) -> List[_PostProcessingJob]:
    jobs: List[_PostProcessingJob] = []
    for module_name in modules or []:
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            pass
        else:
            if hasattr(module, "post_processing"):
                print(f"Adding post-processing job from {module_name}")
                jobs.append(getattr(module, "post_processing"))
    return jobs


def _load_preprocessing_job(module_name: Optional[str]) -> Optional[_PreProcessingJob]:
    if not module_name:
        return None

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        pass
    else:
        if hasattr(module, "pre_processing"):
            print(f"Adding pre-processing job from {module_name}")
            return getattr(module, "pre_processing")
    return None


def _select_stocks(finder: AiZynthFinder, args: argparse.Namespace) -> None:
    stocks = list(args.stocks)
    try:
        module = importlib.import_module("custom_stock")
    except ModuleNotFoundError:
        pass
    else:
        if hasattr(module, "stock"):
            finder.stock.load(module.stock, "custom_stock")  # type: ignore
            stocks.append("custom_stock")
    finder.stock.select(stocks or finder.stock.items)


def _load_checkpoint(
    checkpoint: str,
) -> Dict[str, List[Any]]:
    if not os.path.exists(checkpoint):
        return defaultdict(list)
    with open(checkpoint) as json_file:
        checkpoint_data = [json.loads(line) for line in json_file]

    checkpoint_results = defaultdict(list)
    if checkpoint_data:
        checkpoint_results["processed_smiles"] = [
            data["processed_smiles"] for data in checkpoint_data
        ]
        for data in checkpoint_data:
            for key, value in data["results"].items():
                checkpoint_results[key].append(value)
    return checkpoint_results


def _process_single_smiles(
    smiles: str,
    finder: AiZynthFinder,
    output_name: str,
    do_clustering: bool,
    route_distance_model: Optional[str],
    post_processing: List[_PostProcessingJob],
    pre_processing: Optional[_PreProcessingJob],
) -> None:
    output_name = output_name or "trees.json"
    finder.target_smiles = smiles
    if pre_processing:
        pre_processing(finder, -1)
    try:
        finder.prepare_tree()
    except ValueError as err:
        print(f"Failed to setup search due to: '{str(err).lower()}'")
        return
    finder.tree_search(show_progress=True)
    finder.build_routes()
    finder.routes.compute_scores(*finder.scorers.objects())

    with open(output_name, "w") as fileobj:
        json.dump(
            finder.routes.dict_with_extra(include_metadata=True, include_scores=True),
            fileobj,
            indent=2,
        )
    logger().info(f"Trees saved to {output_name}")

    stats = finder.extract_statistics()
    if do_clustering:
        _do_clustering(
            finder, stats, detailed_results=False, model_path=route_distance_model
        )
    _do_post_processing(finder, stats, post_processing)
    stats_str = "\n".join(
        f"{key.replace('_', ' ')}: {value}" for key, value in stats.items()
    )
    logger().info(stats_str)


def _process_multi_smiles(
    filename: str,
    finder: AiZynthFinder,
    output_name: str,
    do_clustering: bool,
    route_distance_model: Optional[str],
    post_processing: List[_PostProcessingJob],
    pre_processing: Optional[_PreProcessingJob],
    checkpoint: Optional[str],
) -> None:
    output_name = output_name or "output.json.gz"
    with open(filename, "r") as fileobj:
        smiles = [line.strip() for line in fileobj.readlines()]

    checkpoint_data: StrDict = defaultdict(list)
    if checkpoint:
        checkpoint_data = _load_checkpoint(checkpoint)
        start = len(checkpoint_data["processed_smiles"]) if checkpoint_data else 0
        smiles = smiles[start:]

    results: StrDict = defaultdict(list)
    if checkpoint_data:
        results = {
            key: value
            for key, value in checkpoint_data.items()
            if key != "processed_smiles"
        }
    for idx, smi in enumerate(smiles):
        if pre_processing:
            pre_processing(finder, idx)
        processed_results = {}
        finder.target_smiles = smi
        try:
            finder.prepare_tree()
        except ValueError as err:
            print(f"Failed to setup search for {smi} due to: '{str(err).lower()}'")
            continue
        search_time = finder.tree_search()
        finder.build_routes()
        finder.routes.compute_scores(*finder.scorers.objects())
        stats = finder.extract_statistics()

        solved_str = "is solved" if stats["is_solved"] else "is not solved"
        logger().info(f"Done with {smi} in {search_time:.3} s and {solved_str}")
        if do_clustering:
            _do_clustering(
                finder, stats, detailed_results=True, model_path=route_distance_model
            )
        _do_post_processing(finder, stats, post_processing)

        for key, value in stats.items():
            processed_results[key] = value
        processed_results["stock_info"] = finder.stock_info()
        processed_results["trees"] = finder.routes.dict_with_extra(
            include_metadata=True, include_scores=True
        )

        if checkpoint:
            with open(checkpoint, "a") as checkpoint_file:
                checkpoint_file.write(
                    json.dumps({"processed_smiles": smi, "results": processed_results})
                    + "\n"
                )
            logger().debug(
                f"Results for processed smiles '{smi}' saved to {checkpoint}"
            )

        for key, value in processed_results.items():
            results[key].append(value)

    data = pd.DataFrame.from_dict(results)
    save_datafile(data, output_name)
    logger().info(f"Output saved to {output_name}")


def _multiprocess_smiles(args: argparse.Namespace) -> None:
    def create_cmd(index, filename):
        cmd_args = [
            "aizynthcli",
            "--smiles",
            filename,
            "--config",
            args.config,
            "--output",
            json_files[index - 1],
        ]
        if args.policy:
            cmd_args.extend(["--policy"] + args.policy)
        if args.filter:
            cmd_args.extend(["--filter"] + args.filter)
        if args.stocks:
            cmd_args.append("--stocks")
            cmd_args.extend(args.stocks)
        if args.cluster:
            cmd_args.append("--cluster")
        if args.route_distance_model:
            cmd_args.extend(["--route_distance_model", args.route_distance_model])
        if args.post_processing:
            cmd_args.extend(["--post_processing"] + args.post_processing)
        return cmd_args

    if not os.path.exists(args.smiles):
        raise ValueError(
            "For multiprocessing execution the --smiles argument needs to be a filename"
        )

    setup_logger(logging.INFO)
    filenames = split_file(args.smiles, args.nproc)
    json_files = [tempfile.mktemp(suffix=".json.gz") for _ in range(args.nproc)]
    start_processes(filenames, "aizynthcli", create_cmd)

    if not all(os.path.exists(filename) for filename in json_files):
        raise FileNotFoundError(
            "Not all output files produced. Please check the individual log files: 'aizynthcli*.log'"
        )
    cat_datafiles(json_files, args.output or "output.json.gz")


def main() -> None:
    """Entry point for the aizynthcli command"""
    args = _get_arguments()

    file_level_logging = logging.DEBUG if args.log_to_file else None
    setup_logger(logging.INFO, file_level_logging)

    if not os.path.exists(args.smiles):
        mol = Molecule(smiles=args.smiles)
        if mol.rd_mol is None:
            logger().error(
                f"The --smiles argument ({args.smiles})"
                " does not point to an existing file or is a valid RDKit SMILES."
                " Cannot start retrosynthesis planning."
            )
            return

    if args.nproc:
        _multiprocess_smiles(args)
        return

    multi_smiles = os.path.exists(args.smiles)

    finder = AiZynthFinder(configfile=args.config)
    _select_stocks(finder, args)
    post_processing = _load_postprocessing_jobs(args.post_processing)
    pre_processing = _load_preprocessing_job(args.pre_processing)
    finder.expansion_policy.select(args.policy or finder.expansion_policy.items[0])
    if args.filter:
        finder.filter_policy.select(args.filter)
    else:
        finder.filter_policy.select_all()

    params = [
        args.smiles,
        finder,
        args.output,
        args.cluster,
        args.route_distance_model,
        post_processing,
        pre_processing,
        args.checkpoint,
    ]
    if multi_smiles:
        _process_multi_smiles(*params)
    else:
        params = params[:-1]
        _process_single_smiles(*params)


if __name__ == "__main__":
    main()



""" Module containing routines to obtain a MongoClient instance
"""
from typing import Optional
from urllib.parse import urlencode

try:
    from pymongo import MongoClient
except ImportError:
    MongoClient = None
    HAS_PYMONGO = False
else:
    HAS_PYMONGO = True

from aizynthfinder.utils.logging import logger

_CLIENT = None


def get_mongo_client(
    host: str = "localhost",
    port: int = 27017,
    user: Optional[str] = None,
    password: Optional[str] = None,
    tls_certs_path: str = "",
) -> Optional[MongoClient]:
    """
    A helper function to create and reuse MongoClient

    The client is only setup once. Therefore if this function is called a second
    time with different parameters, it would still return the first client.

    :param host: the host
    :param port: the host port
    :param user: username, defaults to None
    :param password: password, defaults to None
    :param tls_certs_path: the path to TLS certificates if to be used, defaults to ""
    :raises ValueError: if host and port is not given first time
    :return: the MongoDB client
    """
    if not HAS_PYMONGO:
        return None

    global _CLIENT
    if _CLIENT is None:
        params = {}
        if tls_certs_path:
            params.update({"ssl": "true", "ssl_ca_certs": tls_certs_path})
        cred_str = f"{user}:{password}@" if password else ""
        uri = f"mongodb://{cred_str}{host}:{port}/?{urlencode(params)}"
        logger().debug(f"Connecting to MongoDB on {host}:{port}")
        _CLIENT = MongoClient(uri)  # pylint: disable=C0103
    return _CLIENT



""" Module containing Retro* cost models """
from __future__ import annotations

import pickle
from typing import TYPE_CHECKING

import numpy as np

from aizynthfinder.search.retrostar.cost import __name__ as retrostar_cost_module
from aizynthfinder.utils.loading import load_dynamic_class

if TYPE_CHECKING:
    from aizynthfinder.chem import Molecule
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import Any, List, Tuple


class MoleculeCost:
    """
    A class to compute the molecule cost.

    The cost to be computed is taken from the input config. If no `molecule_cost` is
    set, assigns ZeroMoleculeCost as the `cost` by default. The `molecule_cost` can be
    set as a dictionary in config under `search` in the following format:
    'algorithm': 'retrostar'
    'algorithm_config': {
        'molecule_cost': {
            'cost': name of the search cost class or custom_package.custom_model.CustomClass,
            other settings or params
        }
    }

    The cost can be computed by calling the instantiated class with a molecule.

    .. code-block::

        calculator = MyCost(config)
        cost = calculator.calculate(molecule)

    :param config: the configuration of the tree search
    """

    def __init__(self, config: Configuration) -> None:
        self._config = config
        if "molecule_cost" not in self._config.search.algorithm_config:
            self._config.search.algorithm_config["molecule_cost"] = {
                "cost": "ZeroMoleculeCost"
            }
        kwargs = self._config.search.algorithm_config["molecule_cost"].copy()

        cls = load_dynamic_class(kwargs["cost"], retrostar_cost_module)
        del kwargs["cost"]

        self.molecule_cost = cls(**kwargs) if kwargs else cls()

    def __call__(self, mol: Molecule) -> float:
        return self.molecule_cost.calculate(mol)


class RetroStarCost:
    """
    Encapsulation of the original Retro* molecular cost model

    Numpy implementation of original pytorch model

    The predictions of the score is made on a Molecule object

    .. code-block::

        mol = Molecule(smiles="CCC")
        scorer = RetroStarCost()
        score = scorer.calculate(mol)

    The model provided when creating the scorer object should be a pickled
    tuple.
    The first item of the tuple should be a list of the model weights for each layer.
    The second item of the tuple should be a list of the model biases for each layer.

    :param model_path: the filename of the model weights and biases
    :param fingerprint_length: the number of bits in the fingerprint
    :param fingerprint_radius: the radius of the fingerprint
    :param dropout_rate: the dropout_rate
    """

    _required_kwargs = ["model_path"]

    def __init__(self, **kwargs: Any) -> None:
        model_path = kwargs["model_path"]
        self.fingerprint_length: int = int(kwargs.get("fingerprint_length", 2048))
        self.fingerprint_radius: int = int(kwargs.get("fingerprint_radius", 2))
        self.dropout_rate: float = float(kwargs.get("dropout_rate", 0.1))

        self._dropout_prob = 1.0 - self.dropout_rate
        self._weights, self._biases = self._load_model(model_path)

    def __repr__(self) -> str:
        return "retrostar"

    def calculate(self, mol: Molecule) -> float:
        # pylint: disable=invalid-name
        mol.sanitize()
        vec = mol.fingerprint(
            radius=self.fingerprint_radius, nbits=self.fingerprint_length
        )
        for W, b in zip(self._weights[:-1], self._biases[:-1]):
            vec = np.matmul(vec, W) + b
            vec *= vec > 0  # ReLU
            # Drop-out
            vec *= np.random.binomial(1, self._dropout_prob, size=vec.shape) / (
                self._dropout_prob
            )
        vec = np.matmul(vec, self._weights[-1]) + self._biases[-1]
        return float(np.log(1 + np.exp(vec)))

    @staticmethod
    def _load_model(model_path: str) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        with open(model_path, "rb") as fileobj:
            weights, biases = pickle.load(fileobj)

        return (
            [np.asarray(item) for item in weights],
            [np.asarray(item) for item in biases],
        )


class ZeroMoleculeCost:
    """Encapsulation of a Zero cost model"""

    def __repr__(self) -> str:
        return "zero"

    def calculate(self, _mol: Molecule) -> float:  # pytest: disable=unused-argument
        return 0.0



""" Module containing all types and type imports
"""
# pylint: disable=unused-import
from typing import Callable  # noqa
from typing import Iterable  # noqa
from typing import List  # noqa
from typing import Sequence  # noqa
from typing import Set  # noqa
from typing import TypeVar  # noqa
from typing import Any, Dict, Optional, Tuple, Union

from PIL.Image import Image
from rdkit import Chem
from rdkit.Chem import rdChemReactions
from rdkit.DataStructs.cDataStructs import ExplicitBitVect

StrDict = Dict[str, Any]
RdMol = Chem.rdchem.Mol
RdReaction = Chem.rdChemReactions.ChemicalReaction
BitVector = ExplicitBitVect
PilImage = Image
PilColor = Union[str, Tuple[int, int, int]]
FrameColors = Optional[Dict[bool, PilColor]]



""" Module containing classes that implements different filter policy strategies
"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

import numpy as np

from aizynthfinder.chem import TemplatedRetroReaction
from aizynthfinder.context.policy.utils import _make_fingerprint
from aizynthfinder.utils.bonds import BrokenBonds
from aizynthfinder.utils.exceptions import (
    PolicyException,
    RejectionException,
)
from aizynthfinder.utils.logging import logger
from aizynthfinder.utils.models import load_model

if TYPE_CHECKING:
    from aizynthfinder.chem.reaction import RetroReaction
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import Any, Dict, List, Optional, Tuple


class FilterStrategy(abc.ABC):
    """
    A base class for all filter strategies.

    The filter can be applied by either calling the `apply` method
    of by calling the instantiated class with a reaction.

    .. code-block::

        filter = MyFilterStrategy("dummy", config)
        filter.apply(reaction)
        filter(reaction)

    :param key: the key or label
    :param config: the configuration of the tree search
    """

    _required_kwargs: List[str] = []

    def __init__(self, key: str, config: Configuration, **kwargs: Any) -> None:
        if any(name not in kwargs for name in self._required_kwargs):
            raise PolicyException(
                f"A {self.__class__.__name__} class needs to be initiated "
                f"with keyword arguments: {', '.join(self._required_kwargs)}"
            )
        self._config = config
        self._logger = logger()
        self.key = key

    def __call__(self, reaction: RetroReaction) -> None:
        self.apply(reaction)

    @abc.abstractmethod
    def apply(self, reaction: RetroReaction) -> None:
        """
        Apply the filter on the reaction. If the reaction
        should be rejected a `RejectionException` is raised

        :param reaction: the reaction to filter
        :raises: if the reaction should be rejected.
        """


class BondFilter(FilterStrategy):
    """
    Check if focussed bonds to freeze stay frozen in a reaction.

    :param key: the key or label
    :param config: the configuration of the tree search
    """

    def __init__(self, key: str, config: Configuration, **kwargs: Any) -> None:
        super().__init__(key, config, **kwargs)

        self._freeze_bonds = config.search.freeze_bonds
        self._broken_bonds = BrokenBonds(self._freeze_bonds)
        self._logger.info(
            f"Loading bond filter to {key} with {len(self._freeze_bonds)} "
            "bonds to freeze"
        )

    def apply(self, reaction: RetroReaction) -> None:
        broken_frozen_bonds = self._broken_bonds(reaction)
        if len(broken_frozen_bonds) > 0:
            raise RejectionException(
                f"{reaction} was filtered out as the focussed bonds "
                f"'{broken_frozen_bonds}' were found to be broken in the reaction"
            )


class QuickKerasFilter(FilterStrategy):
    """
    Filter quick-filter trained on artificial negative data

    :ivar use_remote_models: a boolean to connect to remote TensorFlow servers. Defaults
        to False.
    :ivar filter_cutoff: the cut-off value

    :param key: the key or label
    :param config: the configuration of the tree search
    :param model: the source of the policy model
    """

    _required_kwargs: List[str] = ["model"]

    def __init__(self, key: str, config: Configuration, **kwargs: Any) -> None:
        super().__init__(key, config, **kwargs)
        source = kwargs["model"]
        # self.settings = self._config.filter_settings
        self._logger.info(f"Loading filter policy model from {source} to {key}")
        self.use_remote_models: bool = bool(kwargs.get("use_remote_models", False))
        self.model = load_model(source, key, self.use_remote_models)
        self._prod_fp_name = kwargs.get("prod_fp_name", "input_1")
        self._rxn_fp_name = kwargs.get("rxn_fp_name", "input_2")
        self._exclude_from_policy: List[str] = kwargs.get("exclude_from_policy", [])
        self.filter_cutoff: float = float(kwargs.get("filter_cutoff", 0.05))

    def apply(self, reaction: RetroReaction) -> None:
        if reaction.metadata.get("policy_name", "") in self._exclude_from_policy:
            return

        feasible, prob = self.feasibility(reaction)
        if not feasible:
            raise RejectionException(f"{reaction} was filtered out with prob {prob}")

    def feasibility(self, reaction: RetroReaction) -> Tuple[bool, float]:
        """
        Computes if a given reaction is feasible by given
        the reaction fingerprint to a network model

        :param reaction: the reaction to query
        :return: if the reaction is feasible
        """
        if not reaction.reactants:
            return False, 0.0

        prob = self._predict(reaction)
        feasible = prob >= self.filter_cutoff
        return feasible, prob

    def _predict(self, reaction: RetroReaction) -> float:
        prod_fp, rxn_fp = self._reaction_to_fingerprint(reaction, self.model)
        kwargs = {self._prod_fp_name: prod_fp, self._rxn_fp_name: rxn_fp}
        return self.model.predict(prod_fp, rxn_fp, **kwargs)[0][0]

    @staticmethod
    def _reaction_to_fingerprint(
        reaction: RetroReaction, model: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        rxn_fp = _make_fingerprint(reaction, model)
        prod_fp = _make_fingerprint(reaction.mol, model)
        return prod_fp, rxn_fp


class ReactantsCountFilter(FilterStrategy):
    """
    Check that the number of reactants is was expected from the template

    :param key: the key or label
    :param config: the configuration of the tree search
    """

    def __init__(self, key: str, config: Configuration, **kwargs: Any) -> None:
        super().__init__(key, config, **kwargs)
        self._logger.info(f"Loading reactants count filter to {key}")

    def apply(self, reaction: RetroReaction) -> None:
        if not isinstance(reaction, TemplatedRetroReaction):
            raise ValueError(
                "Reactants count filter can only be used on templated retro reaction "
            )

        reactants = reaction.reactants[reaction.index]
        if len(reactants) > reaction.rd_reaction.GetNumProductTemplates():
            raise RejectionException(
                f"{reaction} was filtered out because number of reactants disagree with the template"
            )


FILTER_STRATEGY_ALIAS = {
    "feasibility": "QuickKerasFilter",
    "quick_keras_filter": "QuickKerasFilter",
    "reactants_count": "ReactantsCountFilter",
}



""" Module containing a class that is the base class for all collection classes (stock, policies, scorers)
"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import Any, List, Optional, StrDict, Union


class ContextCollection(abc.ABC):
    """
    Abstract base class for a collection of items
    that can be loaded and then (de-)selected.

     One can obtain individual items with:

    .. code-block::

        an_item = collection["key"]

    And delete items with

    .. code-block::

        del collection["key"]


    """

    _single_selection = False
    _collection_name = "collection"

    def __init__(self) -> None:
        self._items: StrDict = {}
        self._selection: List[str] = []
        self._logger = logger()

    def __delitem__(self, key: str) -> None:
        if key not in self._items:
            raise KeyError(
                f"{self._collection_name.capitalize()} with name {key} not loaded."
            )
        del self._items[key]

    def __getitem__(self, key: str) -> Any:
        if key not in self._items:
            raise KeyError(
                f"{self._collection_name.capitalize()} with name {key} not loaded."
            )
        return self._items[key]

    def __len__(self) -> int:
        return len(self._items)

    @property
    def items(self) -> List[str]:
        """The available item keys"""
        return list(self._items.keys())

    @property
    def selection(self) -> Union[List[str], str, None]:
        """The keys of the selected item(s)"""
        if self._single_selection:
            return self._selection[0] if self._selection else None
        return self._selection

    @selection.setter
    def selection(self, value: str) -> None:
        self.select(value)

    def deselect(self, key: Optional[str] = None) -> None:
        """
        Deselect one or all items

        If no key is passed, all items will be deselected.

        :param key: the key of the item to deselect, defaults to None
        :raises KeyError: if the key is not among the selected ones
        """
        if not key:
            self._selection = []
            return

        if key not in self._selection:
            raise KeyError(f"Cannot deselect {key} because it is not selected")
        self._selection.remove(key)

    @abc.abstractmethod
    def load(self, *_: Any) -> None:
        """Load an item. Needs to be implemented by a sub-class"""

    @abc.abstractmethod
    def load_from_config(self, **config: Any) -> None:
        """Load items from a configuration. Needs to be implemented by a sub-class"""

    def select(self, value: Union[str, List[str]], append: bool = False) -> None:
        """
        Select one or more items.

        If this is a single selection collection, only a single value is accepted.
        If this is a multiple selection collection it will overwrite the selection completely,
        unless ``append`` is True and a single key is given.

        :param value: the key or keys of the item(s) to select
        :param append: if True will append single keys to existing selection
        :raises ValueError: if this a single collection and value is multiple keys
        :raises KeyError: if at least one of the keys are not corresponding to a loaded item
        """
        if self._single_selection and not isinstance(value, str) and len(value) > 1:
            raise ValueError(f"Cannot select more than one {self._collection_name}")

        keys = [value] if isinstance(value, str) else value

        for key in keys:
            if key not in self._items:
                raise KeyError(
                    f"Invalid key specified {key} when selecting {self._collection_name}"
                )

        if self._single_selection:
            self._selection = [keys[0]]
        elif isinstance(value, str) and append:
            self._selection.append(value)
        else:
            self._selection = list(keys)

        self._logger.info(f"Selected as {self._collection_name}: {', '.join(keys)}")

    def select_all(self) -> None:
        """Select all loaded items"""
        if self.items:
            self.select(self.items)

    def select_first(self) -> None:
        """Select the first loaded item"""
        if self.items:
            self.select(self.items[0])

    def select_last(self) -> None:
        """Select the last loaded item"""
        if self.items:
            self.select(self.items[-1])



""" Sub-package containing stock routines
"""
from aizynthfinder.context.stock.queries import (
    InMemoryInchiKeyQuery,
    MongoDbInchiKeyQuery,
    StockQueryMixin,
)
from aizynthfinder.context.stock.stock import Stock
from aizynthfinder.utils.exceptions import StockException
